"""Exceptions used by the fudge module.

See :ref:`using-fudge` for common scenarios.
"""

__all__ = ["FakeDeclarationError"]


class FakeDeclarationError(Exception):
    """Exception in how this :class:`fudge.Fake` was declared."""



import threading
import unittest
import fudge

from fudge import ExpectedCall, ExpectedCallOrder


class TestRegistry(unittest.TestCase):
    def setUp(self):
        self.fake = fudge.Fake()
        self.reg = fudge.registry
        # in case of error, clear out everything:
        self.reg.clear_all()

    def tearDown(self):
        pass

    def test_expected_call_not_called(self):
        with self.assertRaises(AssertionError):
            self.reg.clear_calls()
            self.reg.expect_call(ExpectedCall(self.fake, "nothing"))
            self.reg.verify()

    def test_clear_calls_resets_calls(self):
        exp = ExpectedCall(self.fake, "callMe")
        self.reg.expect_call(exp)
        exp()
        self.assertEqual(exp.was_called, True)

        self.reg.clear_calls()
        self.assertEqual(exp.was_called, False, "call was not reset by clear_calls()")

    def test_clear_calls_resets_call_order(self):
        exp_order = ExpectedCallOrder(self.fake)
        exp = ExpectedCall(self.fake, "callMe", call_order=exp_order)
        exp_order.add_expected_call(exp)
        self.reg.remember_expected_call_order(exp_order)

        exp()
        self.assertEqual(exp_order._actual_calls, [exp])

        self.reg.clear_calls()
        self.assertEqual(
            exp_order._actual_calls, [], "call order calls were not reset by clear_calls()"
        )

    def test_verify_resets_calls(self):
        exp = ExpectedCall(self.fake, "callMe")
        exp()
        self.assertEqual(exp.was_called, True)
        self.assertEqual(len(self.reg.get_expected_calls()), 1)

        self.reg.verify()
        self.assertEqual(exp.was_called, False, "call was not reset by verify()")
        self.assertEqual(
            len(self.reg.get_expected_calls()), 1, "verify() should not reset expectations"
        )

    def test_verify_resets_call_order(self):
        exp_order = ExpectedCallOrder(self.fake)
        exp = ExpectedCall(self.fake, "callMe", call_order=exp_order)
        exp_order.add_expected_call(exp)
        self.reg.remember_expected_call_order(exp_order)

        exp()
        self.assertEqual(exp_order._actual_calls, [exp])

        self.reg.verify()
        self.assertEqual(exp_order._actual_calls, [], "call order calls were not reset by verify()")

    def test_global_verify(self):
        exp = ExpectedCall(self.fake, "callMe")
        exp()
        self.assertEqual(exp.was_called, True)
        self.assertEqual(len(self.reg.get_expected_calls()), 1)

        fudge.verify()

        self.assertEqual(exp.was_called, False, "call was not reset by verify()")
        self.assertEqual(
            len(self.reg.get_expected_calls()), 1, "verify() should not reset expectations"
        )

    def test_global_clear_expectations(self):
        exp = ExpectedCall(self.fake, "callMe")
        exp()
        self.assertEqual(len(self.reg.get_expected_calls()), 1)
        exp_order = ExpectedCallOrder(self.fake)
        self.reg.remember_expected_call_order(exp_order)
        keys = list(self.reg.get_expected_call_order().keys())
        self.assertEqual(keys, [self.fake])

        fudge.clear_expectations()

        self.assertEqual(
            len(self.reg.get_expected_calls()), 0, "clear_expectations() should reset expectations"
        )
        self.assertEqual(
            len(self.reg.get_expected_call_order().keys()),
            0,
            "clear_expectations() should reset expected call order",
        )

    def test_multithreading(self):
        reg = fudge.registry

        class thread_run:
            waiting = 5
            errors = []

        # while this barely catches collisions
        # it ensures that each threading can use the registry ok
        def registry(num):
            try:
                try:
                    fudge.clear_calls()
                    fudge.clear_expectations()

                    exp_order = ExpectedCallOrder(self.fake)
                    reg.remember_expected_call_order(exp_order)
                    self.assertEqual(len(reg.get_expected_call_order().keys()), 1)

                    # registered first time on __init__ :
                    exp = ExpectedCall(self.fake, "callMe", call_order=exp_order)
                    reg.expect_call(exp)
                    reg.expect_call(exp)
                    reg.expect_call(exp)
                    self.assertEqual(len(reg.get_expected_calls()), 4)

                    # actual calls:
                    exp()
                    exp()
                    exp()
                    exp()

                    fudge.verify()
                    fudge.clear_expectations()
                except Exception as exc:
                    thread_run.errors.append(exc)
                    raise
            finally:
                thread_run.waiting -= 1

        threading.Thread(target=registry, args=(1,)).start()
        threading.Thread(target=registry, args=(2,)).start()
        threading.Thread(target=registry, args=(3,)).start()
        threading.Thread(target=registry, args=(4,)).start()
        threading.Thread(target=registry, args=(5,)).start()

        count = 0
        while thread_run.waiting > 0:
            count += 1
            import time

            time.sleep(0.25)
            if count == 60:
                raise RuntimeError("timed out waiting for threading")
        if len(thread_run.errors):
            raise RuntimeError(
                "Error(s) in threading: %s"
                % ["%s: %s" % (e.__class__.__name__, e) for e in thread_run.errors]
            )



import unittest

import fudge
from fudge import inspector
from fudge.inspector import arg
from fudge import Fake


class TestAnyValue(unittest.TestCase):

    def tearDown(self):
        fudge.clear_expectations()

    def test_any_value(self):
        db = Fake("db").expects("execute").with_args(arg.any())
        db.execute("delete from foo where 1")

    def test_repr(self):
        any = inspector.AnyValue()
        self.assertEqual(repr(any), "arg.any()")

    def test_str(self):
        any = inspector.AnyValue()
        self.assertEqual(str(any), "arg.any()")


class TestPassesTest(unittest.TestCase):

    def tearDown(self):
        fudge.clear_expectations()

    def test_passes(self):
        def isint(v):
            return isinstance(v, int)

        counter = Fake("counter").expects("increment").with_args(arg.passes_test(isint))
        counter.increment(25)

    def test_passes_fail(self):
        def is_str(v):
            return isinstance(v, str)

        with self.assertRaises(AssertionError):
            counter = Fake("counter").expects("set_name").with_args(arg.passes_test(is_str))
            counter.set_name(25)

    def test_repr(self):
        class test(object):
            def __call__(self, v):
                return isinstance(v, int)

            def __repr__(self):
                return "v is an int"

        passes = inspector.PassesTest(test())
        self.assertEqual(repr(passes), "arg.passes_test(v is an int)")

    def test_str(self):
        class test(object):
            def __call__(self, v):
                return isinstance(v, int)

            def __repr__(self):
                return "v is an int"

        passes = inspector.PassesTest(test())
        self.assertEqual(str(passes), "arg.passes_test(v is an int)")


class TestIsInstance(unittest.TestCase):

    def tearDown(self):
        fudge.clear_expectations()

    def test_passes(self):
        counter = Fake("counter").expects("increment").with_args(arg.isinstance(int))
        counter.increment(25)

    def test_passes_fail(self):
        with self.assertRaises(AssertionError):
            counter = Fake("counter").expects("set_name").with_args(arg.isinstance(str))
            counter.set_name(25)

    def test_repr(self):
        passes = inspector.IsInstance(int)
        self.assertEqual(repr(passes), "arg.isinstance('int')")

    def test_str(self):
        passes = inspector.IsInstance(str)
        self.assertEqual(str(passes), "arg.isinstance('str')")

    def test_list(self):
        passes = inspector.IsInstance((str, int))
        self.assertEqual(str(passes), "arg.isinstance(('str', 'int'))")


class TestObjectlike(unittest.TestCase):

    def tearDown(self):
        fudge.clear_expectations()

    def test_has_attr_ok(self):
        class Config(object):
            size = 12
            color = "red"
            weight = "heavy"

        widget = Fake("widget").expects("configure").with_args(arg.has_attr(size=12, color="red"))
        widget.configure(Config())

    def test_has_attr_fail(self):
        class Config(object):
            color = "red"

        with self.assertRaises(AssertionError):
            widget = Fake("widget").expects("configure").with_args(arg.has_attr(size=12))
            widget.configure(Config())

    def test_has_attr_fail_wrong_value(self):
        class Config(object):
            color = "red"

        with self.assertRaises(AssertionError):
            widget = Fake("widget").expects("configure").with_args(arg.has_attr(color="green"))
            widget.configure(Config())

    def test_objectlike_str(self):
        o = inspector.HasAttr(one=1, two="two")
        self.assertEqual(str(o), "arg.has_attr(one=1, two='two')")

    def test_objectlike_repr(self):
        o = inspector.HasAttr(one=1, two="two")
        self.assertEqual(repr(o), "arg.has_attr(one=1, two='two')")

    def test_objectlike_unicode(self):
        o = inspector.HasAttr(one=1, ivan="Ivan_Krsti\u0107")
        self.assertEqual(repr(o), "arg.has_attr(ivan=%s, one=1)" % repr("Ivan_Krsti\u0107"))

    def test_objectlike_repr_long_val(self):
        o = inspector.HasAttr(bytes="011110101000101010011111111110000001010100000001110000000011")
        self.assertEqual(
            repr(o), "arg.has_attr(bytes='011110101000101010011111111110000001010100000...')"
        )


class TestStringlike(unittest.TestCase):

    def tearDown(self):
        fudge.clear_expectations()

    def test_startswith_ok(self):
        db = Fake("db").expects("execute").with_args(arg.startswith("insert into"))
        db.execute("insert into foo values (1,2,3,4)")

    def test_startswith_fail(self):
        with self.assertRaises(AssertionError):
            db = Fake("db").expects("execute").with_args(arg.startswith("insert into"))
            db.execute("select from")

    def test_startswith_ok_uni(self):
        db = Fake("db").expects("execute").with_args(arg.startswith("Ivan_Krsti\u0107"))
        db.execute("Ivan_Krsti\u0107(); foo();")

    def test_startswith_unicode(self):
        p = inspector.Startswith("Ivan_Krsti\u0107")
        self.assertEqual(repr(p), "arg.startswith(%s)" % repr("Ivan_Krsti\u0107"))

    def test_endswith_ok(self):
        db = Fake("db").expects("execute").with_args(arg.endswith("values (1,2,3,4)"))
        db.execute("insert into foo values (1,2,3,4)")

    def test_endswith_ok_uni(self):
        db = Fake("db").expects("execute").with_args(arg.endswith("Ivan Krsti\u0107"))
        db.execute("select Ivan Krsti\u0107")

    def test_endswith_unicode(self):
        p = inspector.Endswith("Ivan_Krsti\u0107")
        self.assertEqual(repr(p), "arg.endswith(%s)" % repr("Ivan_Krsti\u0107"))

    def test_startswith_repr(self):
        p = inspector.Startswith("_start")
        self.assertEqual(repr(p), "arg.startswith('_start')")

    def test_endswith_repr(self):
        p = inspector.Endswith("_ending")
        self.assertEqual(repr(p), "arg.endswith('_ending')")

    def test_startswith_str(self):
        p = inspector.Startswith("_start")
        self.assertEqual(str(p), "arg.startswith('_start')")

    def test_endswith_str(self):
        p = inspector.Endswith("_ending")
        self.assertEqual(str(p), "arg.endswith('_ending')")

    def test_startswith_str_long_value(self):
        p = inspector.Startswith(
            "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
        )
        self.assertEqual(
            str(p), "arg.startswith('AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA...')"
        )

    def test_endswith_str_long_value(self):
        p = inspector.Endswith(
            "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
        )
        self.assertEqual(str(p), "arg.endswith('AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA...')")


class TestContains(unittest.TestCase):

    def tearDown(self):
        fudge.clear_expectations()

    def test_contains_str(self):
        db = Fake("db").expects("execute").with_args(arg.contains("table foo"))
        db.execute("select into table foo;")
        db.execute("select * from table foo where bar = 1")
        fudge.verify()

    def test_contains_fail(self):
        with self.assertRaises(AssertionError):
            db = Fake("db").expects("execute").with_args(arg.contains("table foo"))
            db.execute("select into table notyourmama;")
            fudge.verify()

    def test_contains_list(self):
        db = Fake("db").expects("execute_statements").with_args(arg.contains("select * from foo"))
        db.execute_statements(["update foo", "select * from foo", "drop table foo"])
        fudge.verify()

    def test_str(self):
        c = inspector.Contains(":part:")
        self.assertEqual(str(c), "arg.contains(':part:')")

    def test_str_long_val(self):
        c = inspector.Contains(
            "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
        )
        self.assertEqual(str(c), "arg.contains('AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA...')")

    def test_repr(self):
        c = inspector.Contains(":part:")
        self.assertEqual(repr(c), "arg.contains(':part:')")

    def test_unicode(self):
        c = inspector.Contains("Ivan_Krsti\u0107")
        self.assertEqual(repr(c), "arg.contains(%s)" % repr("Ivan_Krsti\u0107"))


class TestMakeValueTest(unittest.TestCase):

    def test_no_invert_just_inits(self):
        vi = inspector.ValueInspector()
        vi.invert_eq = False

        def value_test(*args, **kwargs):
            return "hello, friends"

        ret = vi._make_value_test(value_test, "asdf", foo="bar")
        self.assertEqual(ret, "hello, friends")

    def test_invert_returns_inverter(self):
        vi = inspector.ValueInspector()
        vi.invert_eq = True

        class vt(object):
            def __eq__(self, other):
                return True

        inverter = vi._make_value_test(vt)
        assert isinstance(inverter, vt)
        self.assertEqual(inverter.__eq__("whatever"), False)



"""Value inspectors that can be passed to :func:`fudge.Fake.with_args` for more
expressive argument matching.

As a mnemonic device,
an instance of the :class:`fudge.inspector.ValueInspector` is available as "arg" :

.. doctest::

    >>> import fudge
    >>> from fudge.inspector import arg
    >>> image = fudge.Fake("image").expects("save").with_args(arg.endswith(".jpg"))

In other words, this declares that the first argument to ``image.save()``
should end with the suffix ".jpg"

.. doctest::
    :hide:

    >>> fudge.clear_expectations()

"""

import warnings
from fudge.util import fmt_val, fmt_dict_vals

__all__ = ["arg", "arg_not"]


class ValueInspector(object):
    """Dispatches tests to inspect values."""

    invert_eq = False

    def _make_value_test(self, test_class, *args, **kwargs):
        if not self.invert_eq:
            return test_class(*args, **kwargs)

        class ValueTestInverter(test_class):
            def __repr__(wrapper_self):
                return "(NOT) %s" % test_class.__repr__(wrapper_self)

            def __eq__(wrapper_self, other):
                return not test_class.__eq__(wrapper_self, other)

        return ValueTestInverter(*args, **kwargs)

    def any(self):
        """Match any value.

        This is pretty much just a placeholder for when you
        want to inspect multiple arguments but don't care about
        all of them.

        .. doctest::

            >>> import fudge
            >>> from fudge.inspector import arg
            >>> db = fudge.Fake("db")
            >>> db = db.expects("transaction").with_args(
            ...             "insert", isolation_level=arg.any())
            ...
            >>> db.transaction("insert", isolation_level="lock")
            >>> fudge.verify()

        This also passes:

        .. doctest::
            :hide:

            >>> fudge.clear_calls()

        .. doctest::

            >>> db.transaction("insert", isolation_level="autocommit")
            >>> fudge.verify()

        .. doctest::
            :hide:

            >>> fudge.clear_expectations()

        The arg_not version will not match anything and is probably not very
        useful.

        .. doctest::

            >>> import fudge
            >>> from fudge.inspector import arg_not
            >>> query = fudge.Fake('query').expects_call().with_args(
            ...     arg_not.any()
            ... )
            >>> query('asdf')
            Traceback (most recent call last):
            ...
            AssertionError: fake:query((NOT) arg.any()) was called unexpectedly with args ('asdf')
            >>> query()
            Traceback (most recent call last):
            ...
            AssertionError: fake:query((NOT) arg.any()) was called unexpectedly with args ()

        .. doctest::
            :hide:

            >>> fudge.clear_expectations()

        """
        return self._make_value_test(AnyValue)

    def any_value(self):
        """**DEPRECATED**: use :func:`arg.any() <fudge.inspector.ValueInspector.any>`"""
        warnings.warn("arg.any_value() is deprecated in favor of arg.any()", DeprecationWarning, 3)
        return self.any()

    def contains(self, part):
        """Ensure that a value contains some part.

        This is useful for when you only care that a substring or subelement
        exists in a value.

        .. doctest::

            >>> import fudge
            >>> from fudge.inspector import arg
            >>> addressbook = fudge.Fake().expects("import_").with_args(
            ...                                     arg.contains("Baba Brooks"))
            ...
            >>> addressbook.import_("Bill Brooks; Baba Brooks; Henry Brooks;")
            >>> fudge.verify()

        .. doctest::
            :hide:

            >>> fudge.clear_expectations()

        Since contains() just invokes the __in__() method, checking that a list
        item is present works as expected :

        .. doctest::

            >>> colorpicker = fudge.Fake("colorpicker")
            >>> colorpicker = colorpicker.expects("select").with_args(arg.contains("red"))
            >>> colorpicker.select(["green","red","blue"])
            >>> fudge.verify()

        .. doctest::
            :hide:

            >>> fudge.clear_expectations()

        arg_not.contains matches an argument not containing some element.

        .. doctest::

            >>> from fudge.inspector import arg_not
            >>> colorpicker = colorpicker.expects('select').with_args(arg_not.contains('blue'))
            >>> colorpicker.select('reddish')
            >>> colorpicker.select(['red', 'green'])
            >>> fudge.verify()

            >>> colorpicker.select('blue-green')
            Traceback (most recent call last):
            ...
            AssertionError: fake:colorpicker.select(arg.contains('red'))[0]
            was called unexpectedly with args ('blue-green')
            >>> colorpicker.select(['red', 'blue', 'green'])
            Traceback (most recent call last):
            ...
            AssertionError: fake:colorpicker.select((NOT) arg.contains('blue'))[1]
            was called unexpectedly with args (['red', 'blue', 'green'])

        .. doctest::
            :hide:

            >>> fudge.clear_expectations()

        """
        return self._make_value_test(Contains, part)

    def endswith(self, part):
        """Ensure that a value ends with some part.

        This is useful for when values with dynamic parts that are hard to replicate.

        .. doctest::

            >>> import fudge
            >>> from fudge.inspector import arg
            >>> tmpfile = fudge.Fake("tempfile").expects("mkname").with_args(
            ...                                             arg.endswith(".tmp"))
            ...
            >>> tmpfile.mkname("7AakkkLazUUKHKJgh908JKjlkh.tmp")
            >>> fudge.verify()

        .. doctest::
            :hide:

            >>> fudge.clear_expectations()

        The arg_not version works as expected, matching arguments that do not
        end with the given element.

        .. doctest::

            >>> from fudge.inspector import arg_not
            >>> query = fudge.Fake('query').expects_call().with_args(arg_not.endswith('Ringo'))
            >>> query('John, Paul, George and Steve')
            >>> fudge.verify()

        .. doctest::
            :hide:

            >>> fudge.clear_expectations()

        """
        return self._make_value_test(Endswith, part)

    def has_attr(self, **attributes):
        """Ensure that an object value has at least these attributes.

        This is useful for testing that an object has specific attributes.

        .. doctest::

            >>> import fudge
            >>> from fudge.inspector import arg
            >>> db = fudge.Fake("db").expects("update").with_args(arg.has_attr(
            ...                                                       first_name="Bob",
            ...                                                       last_name="James" ))
            ...
            >>> class User:
            ...     first_name = "Bob"
            ...     last_name = "James"
            ...     job = "jazz musician" # this is ignored
            ...
            >>> db.update(User())
            >>> fudge.verify()

        In case of error, the other object's __repr__ will be invoked:

        .. doctest::
            :hide:

            >>> fudge.clear_calls()

        .. doctest::

            >>> class User:
            ...     first_name = "Bob"
            ...
            ...     def __repr__(self):
            ...         return repr(dict(first_name=self.first_name))
            ...
            >>> db.update(User())
            Traceback (most recent call last):
            ...
            AssertionError: fake:db.update(arg.has_attr(first_name='Bob', last_name='James'))
            was called unexpectedly with args ({'first_name': 'Bob'})

        When called as a method on arg_not, has_attr does the opposite, and
        ensures that the argument does not have the specified attributes.

        .. doctest::

            >>> from fudge.inspector import arg_not
            >>> class User:
            ...     first_name = 'Bob'
            ...     last_name = 'Dobbs'
            >>> query = fudge.Fake('query').expects_call().with_args(
            ...     arg_not.has_attr(first_name='James')
            ... )
            >>> query(User())
            >>> fudge.verify()

        .. doctest::
            :hide:

            >>> fudge.clear_expectations()

        """
        return self._make_value_test(HasAttr, **attributes)

    def passes_test(self, test):
        """Check that a value passes some test.

        For custom assertions you may need to create your own callable
        to inspect and verify a value.

        .. doctest::

            >>> def is_valid(s):
            ...     if s in ('active','deleted'):
            ...         return True
            ...     else:
            ...         return False
            ...
            >>> import fudge
            >>> from fudge.inspector import arg
            >>> system = fudge.Fake("system")
            >>> system = system.expects("set_status").with_args(arg.passes_test(is_valid))
            >>> system.set_status("active")
            >>> fudge.verify()

        .. doctest::
            :hide:

            >>> fudge.clear_calls()

        The callable you pass takes one argument, the value, and should return
        True if it's an acceptable value or False if not.

        .. doctest::

            >>> system.set_status("sleep") # doctest: +ELLIPSIS
            Traceback (most recent call last):
            ...
            AssertionError: fake:system.set_status(arg.passes_test(<function is_valid at...))
            was called unexpectedly with args ('sleep')

        .. doctest::
            :hide:

            >>> fudge.clear_expectations()

        If it makes more sense to perform assertions in your test function then
        be sure to return True :

            >>> def is_valid(s):
            ...     assert s in ('active','deleted'), (
            ...         "Unexpected status value: %s" % s)
            ...     return True
            ...
            >>> import fudge
            >>> from fudge.inspector import arg
            >>> system = fudge.Fake("system")
            >>> system = system.expects("set_status").with_args(arg.passes_test(is_valid))
            >>> system.set_status("sleep")
            Traceback (most recent call last):
            ...
            AssertionError: Unexpected status value: sleep

        .. doctest::
            :hide:

            >>> fudge.clear_expectations()

        Using the inverted version, arg_not.passes_test, asserts that the
        argument does not pass the provided test.

        .. doctest::

            >>> from fudge.inspector import arg_not
            >>> query = fudge.Fake('query').expects_call().with_args(
            ...     arg_not.passes_test(lambda x: x > 10)
            ... )
            >>> query(5)
            >>> fudge.verify()

        .. doctest::
            :hide:

            >>> fudge.clear_expectations()

        """
        return self._make_value_test(PassesTest, test)

    def isinstance(self, cls):
        """Check that a value is instance of specified class.

        .. doctest::

            >>> import fudge
            >>> from fudge.inspector import arg
            >>> system = fudge.Fake("system")
            >>> system = system.expects("set_status").with_args(arg.isinstance(str))
            >>> system.set_status("active")
            >>> fudge.verify()

        .. doctest::
            :hide:

            >>> fudge.clear_calls()

        Should return True if it's allowed class or False if not.

        .. doctest::

            >>> system.set_status(31337) # doctest: +ELLIPSIS
            Traceback (most recent call last):
            ...
            AssertionError: fake:system.set_status(arg.isinstance('str'))
            was called unexpectedly with args (31337)

        .. doctest::
            :hide:

            >>> fudge.clear_expectations()

        """
        return self._make_value_test(IsInstance, cls)

    def startswith(self, part):
        """Ensure that a value starts with some part.

        This is useful for when values with dynamic parts that are hard to replicate.

        .. doctest::

            >>> import fudge
            >>> from fudge.inspector import arg
            >>> keychain = fudge.Fake("keychain").expects("accept_key").with_args(
            ...                                                     arg.startswith("_key"))
            ...
            >>> keychain.accept_key("_key-18657yojgaodfty98618652olkj[oollk]")
            >>> fudge.verify()

        .. doctest::
            :hide:

            >>> fudge.clear_expectations()

        Using arg_not.startswith instead ensures that arguments do not start
        with that part.

        .. doctest::

            >>> from fudge.inspector import arg_not
            >>> query = fudge.Fake('query').expects_call().with_args(
            ...     arg_not.startswith('asdf')
            ... )
            >>> query('qwerty')
            >>> fudge.verify()

        .. doctest::
            :hide:

            >>> fudge.clear_expectations()

        """
        return self._make_value_test(Startswith, part)


class NotValueInspector(ValueInspector):
    """Inherits all the argument methods from ValueInspector, but inverts them
    to expect the opposite. See the ValueInspector method docstrings for
    examples.
    """

    invert_eq = True

    def __call__(self, thing):
        """This will match any value except the argument given.

        .. doctest::

            >>> import fudge
            >>> from fudge.inspector import arg, arg_not
            >>> query = fudge.Fake('query').expects_call().with_args(
            ...     arg.any(),
            ...     arg_not('foobar')
            ... )
            >>> query([1, 2, 3], 'asdf')
            >>> query('asdf', 'foobar')
            Traceback (most recent call last):
            ...
            AssertionError: fake:query(arg.any(), arg_not(foobar))
            was called unexpectedly with args ('asdf', 'foobar')

        .. doctest::
            :hide:

            >>> fudge.clear_expectations()

        """
        return NotValue(thing)


arg = ValueInspector()
arg_not = NotValueInspector()


class ValueTest(object):

    arg_method = None
    __test__ = False  # nose

    def __eq__(self, other):
        raise NotImplementedError()

    def _repr_argspec(self):
        raise NotImplementedError()

    def __str__(self):
        return self._repr_argspec()

    def __unicode__(self):
        return self._repr_argspec()

    def __repr__(self):
        return self._repr_argspec()

    def _make_argspec(self, arg):
        if self.arg_method is None:
            raise NotImplementedError("%r must have set attribute arg_method" % self.__class__)
        return "arg." + self.arg_method + "(" + arg + ")"


class Stringlike(ValueTest):

    def __init__(self, part):
        self.part = part

    def _repr_argspec(self):
        return self._make_argspec(fmt_val(self.part))

    def stringlike(self, value):
        if isinstance(value, str):
            return value
        else:
            return str(value)

    def __eq__(self, other):
        check_stringlike = getattr(self.stringlike(other), self.arg_method)
        return check_stringlike(self.part)


class Startswith(Stringlike):
    arg_method = "startswith"


class Endswith(Stringlike):
    arg_method = "endswith"


class HasAttr(ValueTest):
    arg_method = "has_attr"

    def __init__(self, **attributes):
        self.attributes = attributes

    def _repr_argspec(self):
        return self._make_argspec(", ".join(sorted(fmt_dict_vals(self.attributes))))

    def __eq__(self, other):
        for name, value in self.attributes.items():
            if not hasattr(other, name):
                return False
            if getattr(other, name) != value:
                return False

        return True


class AnyValue(ValueTest):
    arg_method = "any"

    def __eq__(self, other):
        # will match anything:
        return True

    def _repr_argspec(self):
        return self._make_argspec("")


class Contains(ValueTest):
    arg_method = "contains"

    def __init__(self, part):
        self.part = part

    def _repr_argspec(self):
        return self._make_argspec(fmt_val(self.part))

    def __eq__(self, other):
        if self.part in other:
            return True
        else:
            return False


class PassesTest(ValueTest):
    arg_method = "passes_test"

    def __init__(self, test):
        self.test = test

    def __eq__(self, other):
        return self.test(other)

    def _repr_argspec(self):
        return self._make_argspec(repr(self.test))


class IsInstance(ValueTest):
    arg_method = "isinstance"

    def __init__(self, cls):
        self.cls = cls

    def __eq__(self, other):
        return isinstance(other, self.cls)

    def _repr_argspec(self):
        if isinstance(self.cls, (tuple, list)):
            return self._make_argspec(repr(tuple((cls.__name__ for cls in self.cls))))
        else:
            return self._make_argspec(repr(self.cls.__name__))


class NotValue(ValueTest):
    def __init__(self, item):
        self.item = item

    def __eq__(self, other):
        return not self.item == other

    def _repr_argspec(self):
        return "arg_not(%s)" % self.item



"""Patching utilities for working with fake objects.

See :ref:`using-fudge` for common scenarios.
"""

__all__ = ["patch_object", "with_patched_object", "PatchHandler", "patched_context", "patch"]

from functools import wraps
import sys
import fudge


class patch(object):
    """A test decorator that patches importable names with :class:`fakes <Fake>`

    Each fake is exposed as an argument to the test:

    .. doctest::
        :hide:

        >>> import fudge

    .. doctest::

        >>> @fudge.patch('os.remove')
        ... def test(fake_remove):
        ...     fake_remove.expects_call()
        ...     # do stuff...
        ...
        >>> test()
        Traceback (most recent call last):
        ...
        AssertionError: fake:os.remove() was not called

    .. doctest::
        :hide:

        >>> fudge.clear_expectations()

    Many paths can be patched at once:

    .. doctest::

        >>> @fudge.patch('os.remove',
        ...              'shutil.rmtree')
        ... def test(fake_remove, fake_rmtree):
        ...     fake_remove.is_callable()
        ...     # do stuff...
        ...

    You can also use :func:`fudge.patch` as a context manager. This is useful when
    working with testing frameworks which mangle the calling signature of test functions,
    such as pytest.

    .. doctest::

        >>> def test():
        ...     with fudge.patch('os.remove') as fake_remove:
        ...         fake_remove.expects_call()
        ...         # do stuff...
        ...
        >>> def test():
        ...     with fudge.patch('os.remove',
        ...                      'shutil.rmtree') as (fake_remove, fake_rmtree):
        ...         fake_remove.is_callable()
        ...         # do stuff...
        ...

    For convenience, the patch method calls
    :func:`fudge.clear_calls`, :func:`fudge.verify`, and :func:`fudge.clear_expectations`.
    For that reason, you must manage all your fake objects within the test function itself.

    .. note::

        If you are using a unittest class, you cannot declare fakes
        within ``setUp()`` unless you manually clear calls and clear
        expectations.  If you do that, you'll want to use the
        :func:`fudge.with_fakes` decorator instead of ``@patch``.

    """

    def __init__(self, *obj_paths):
        self.obj_paths = obj_paths

    def __call__(self, fn):
        @wraps(fn)
        def caller(*args, **kw):
            fakes = self.__enter__()
            if not isinstance(fakes, (tuple, list)):
                fakes = [fakes]
            args += tuple(fakes)
            value = None
            try:
                value = fn(*args, **kw)
            except Exception:
                etype, val, tb = sys.exc_info()
                self.__exit__(etype, val, tb)
                raise
            else:
                self.__exit__(None, None, None)
            return value

        return caller

    def __enter__(self):
        fudge.clear_expectations()
        fudge.clear_calls()
        self.patches = []
        all_fakes = []
        for path in self.obj_paths:
            try:
                target, attr = path.rsplit(".", 1)
            except (TypeError, ValueError):
                raise TypeError("Need a valid target to patch. You supplied: %r" % path)
            fake = fudge.Fake(path)
            all_fakes.append(fake)
            self.patches.append(patch_object(target, attr, fake))
        if len(all_fakes) == 1:
            return all_fakes[0]
        else:
            return all_fakes

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            if not exc_type:
                fudge.verify()
        finally:
            for p in self.patches:
                p.restore()
            fudge.clear_expectations()


def with_patched_object(obj, attr_name, patched_value):
    """Decorator that patches an object before the decorated method
    is called and restores it afterwards.

    This is a wrapper around :func:`fudge.patcher.patch_object`

    Example::

        >>> from fudge import with_patched_object
        >>> class Session:
        ...     state = 'clean'
        ...
        >>> @with_patched_object(Session, "state", "dirty")
        ... def test():
        ...     print Session.state
        ...
        >>> test()
        dirty
        >>> print Session.state
        clean

    """

    def patcher(method):
        @wraps(method)
        def method_call(*m_args, **m_kw):
            patched_obj = patch_object(obj, attr_name, patched_value)
            try:
                return method(*m_args, **m_kw)
            finally:
                patched_obj.restore()

        return method_call

    return patcher


class patched_context(object):
    """A context manager to patch an object temporarily during a `with statement`_ block.

    This is a wrapper around :func:`fudge.patcher.patch_object`

    .. lame, lame, cannot figure out how to apply __future__ to doctest
       so this output is currently skipped

    .. doctest:: python25
        :options: +SKIP

        >>> from fudge import patched_context
        >>> class Session:
        ...     state = 'clean'
        ...
        >>> with patched_context(Session, "state", "dirty"): # doctest: +SKIP
        ...     print Session.state
        ...
        dirty
        >>> print Session.state
        clean

    .. _with statement: http://www.python.org/dev/peps/pep-0343/

    """

    def __init__(self, obj, attr_name, patched_value):

        # note that a @contextmanager decorator would be simpler
        # but it can't be used since a value cannot be yielded within a
        # try/finally block which is needed to restore the object on finally.

        self.patched_object = patch_object(obj, attr_name, patched_value)

    def __enter__(self):
        return self.patched_object

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.patched_object.restore()


def patch_object(obj, attr_name, patched_value):
    """
    Patches an object and returns an instance of :class:`fudge.patcher.PatchHandler`
    for later restoration.

    Note that if *obj* is not an object but a path to a module then it will be imported.

    You may want to use a more convenient wrapper :func:`with_patched_object`
    or :func:`patched_context`

    Example::

        >>> from fudge import patch_object
        >>> class Session:
        ...     state = 'clean'
        ...
        >>> patched_session = patch_object(Session, "state", "dirty")
        >>> Session.state
        'dirty'
        >>> patched_session.restore()
        >>> Session.state
        'clean'

    Here is another example showing how to patch multiple objects at once::

        >>> class Session:
        ...     state = 'clean'
        ...
        >>> class config:
        ...     session_strategy = 'database'
        ...
        >>> patches = [
        ...     patch_object(config, "session_strategy", "filesystem"),
        ...     patch_object(Session, "state", "dirty")
        ... ]
        >>> try:
        ...     # your app under test would run here ...
        ...     print "(while patched)"
        ...     print "config.session_strategy=%r" % config.session_strategy
        ...     print "Session.state=%r" % Session.state
        ... finally:
        ...     for p in patches:
        ...         p.restore()
        ...     print "(patches restored)"
        (while patched)
        config.session_strategy='filesystem'
        Session.state='dirty'
        (patches restored)
        >>> config.session_strategy
        'database'
        >>> Session.state
        'clean'

    """
    if isinstance(obj, str):
        obj_path = adjusted_path = obj
        done = False
        exc = None
        at_top_level = False
        while not done:
            try:
                obj = __import__(adjusted_path)
                done = True
            except ImportError:
                # Handle paths that traveerse object attributes.
                # Such as: smtplib.SMTP.connect
                #          smtplib <- module to import
                adjusted_path = adjusted_path.rsplit(".", 1)[0]
                if not exc:
                    exc = sys.exc_info()
                if at_top_level:
                    # We're at the top level module and it doesn't exist.
                    # Raise the first exception since it will make more sense:
                    raise
                if not adjusted_path.count("."):
                    at_top_level = True
        for part in obj_path.split(".")[1:]:
            obj = getattr(obj, part)

    handle = PatchHandler(obj, attr_name)
    handle.patch(patched_value)
    return handle


class NonExistant(object):
    """Represents a non-existant value."""


class PatchHandler(object):
    """Low level patch handler that memorizes a patch so you can restore it later.

    You can use more convenient wrappers :func:`with_patched_object` and :func:`patched_context`

    """

    def __init__(self, orig_object, attr_name):
        self.orig_object = orig_object
        self.attr_name = attr_name
        self.proxy_object = None
        self.orig_value, self.is_local = self._get_original(self.orig_object, self.attr_name)
        self.getter_class, self.getter = self._handle_getter(self.orig_object, self.attr_name)

    def patch(self, patched_value):
        """Set a new value for the attribute of the object."""
        try:
            if self.getter:
                setattr(self.getter_class, self.attr_name, patched_value)
            else:
                setattr(self.orig_object, self.attr_name, patched_value)
        except TypeError:
            # Workaround for patching builtin objects:
            proxy_name = "fudge_proxy_%s_%s_%s" % (
                self.orig_object.__module__,
                self.orig_object.__name__,
                patched_value.__class__.__name__,
            )
            self.proxy_object = type(
                proxy_name, (self.orig_object,), {self.attr_name: patched_value}
            )
            mod = sys.modules[self.orig_object.__module__]
            setattr(mod, self.orig_object.__name__, self.proxy_object)

    def restore(self):
        """Restore the saved value for the attribute of the object."""
        if self.proxy_object is None:
            if self.getter:
                setattr(self.getter_class, self.attr_name, self.getter)
            elif self.is_local:
                setattr(self.orig_object, self.attr_name, self.orig_value)
            else:
                # Was not a local, safe to delete:
                delattr(self.orig_object, self.attr_name)
        else:
            setattr(
                sys.modules[self.orig_object.__module__],
                self.orig_object.__name__,
                self.orig_object,
            )

    def _find_class_for_attr(self, cls, attr):
        if attr in cls.__dict__:
            return cls
        else:
            for base in cls.__bases__:
                if self._find_class_for_attr(base, attr) is not NonExistant:
                    return base
            return NonExistant

    def _get_original(self, orig_object, name):
        try:
            value = orig_object.__dict__[name]
            is_local = True
        except (AttributeError, KeyError):
            value = getattr(orig_object, name, NonExistant)
            is_local = False
        if value is NonExistant:
            raise AttributeError("%s does not have the attribute %r" % (orig_object, name))
        return value, is_local

    def _get_exact_original(self, orig_object, name):
        if hasattr(orig_object, "__dict__"):
            if name not in orig_object.__dict__:
                # TODO: handle class objects, not just instance objects?
                # This is only here for Class.property.__get__
                if hasattr(orig_object, "__class__"):
                    cls = orig_object.__class__
                    orig_object = self._find_class_for_attr(cls, name)
        return orig_object

    def _handle_getter(self, orig_object, name):
        getter_class, getter = None, None
        exact_orig = self._get_exact_original(orig_object, name)
        try:
            ob = exact_orig.__dict__[name]
        except (AttributeError, KeyError):
            pass
        else:
            if hasattr(ob, "__get__"):
                getter_class = exact_orig
                getter = ob
        return getter_class, getter



def fmt_val(val, shorten=True):
    """Format a value for inclusion in aninformative text string."""
    val = repr(val)
    max = 50
    if shorten:
        if len(val) > max:
            close = val[-1]
            stop = max - 4
            val = val[0:stop] + "..."
            if close in (">", "'", '"', "]", "}", ")"):
                val = val + close
    return val


def fmt_dict_vals(dict_vals, shorten=True):
    """
    Returns list of key=val pairs formatted
    for inclusion in an informative text string.
    """
    items = dict_vals.items()
    if not items:
        return [fmt_val(None, shorten=shorten)]
    return ["%s=%s" % (k, fmt_val(v, shorten=shorten)) for k, v in items]



import unittest

import fudge
from fudge.inspector import arg
from fudge import (
    Fake,
    Registry,
    ExpectedCall,
    ExpectedCallOrder,
    Call,
    CallStack,
    FakeDeclarationError,
)

# for a test below
_some_fake = fudge.Fake()


class TestFake(unittest.TestCase):

    def test_guess_name(self):
        my_obj = fudge.Fake()
        self.assertEqual(repr(my_obj), "fake:my_obj")

    def test_guess_name_globals(self):
        self.assertEqual(repr(_some_fake), "fake:_some_fake")

    def test_guess_name_deref(self):
        my_obj = 44
        my_obj = fudge.Fake()
        self.assertEqual(repr(my_obj), "fake:my_obj")

    def test_has_attr(self):
        my_obj = fudge.Fake().has_attr(vice="versa", beach="playa")
        self.assertEqual(my_obj.vice, "versa")
        self.assertEqual(my_obj.beach, "playa")

    def test_has_property(self):
        fake_vise = fudge.Fake().is_callable().returns("versa")
        fake_stuff = fudge.Fake().is_callable().raises(Exception("broken stuff"))
        my_obj = fudge.Fake().has_property(vice=fake_vise, stuff=fake_stuff)
        self.assertEqual(my_obj.vice, "versa")
        try:
            my_obj.stuff
        except Exception as exc:
            self.assertEqual(str(exc), "broken stuff")
        else:
            raise RuntimeError("expected Exception")

    def test_attributes_are_settable(self):
        my_obj = fudge.Fake().has_attr(vice="versa")
        my_obj.vice = "miami"
        self.assertEqual(my_obj.vice, "miami")

    def test_none_type_attributes_are_settable(self):
        my_obj = fudge.Fake().has_attr(vice=None)
        self.assertEqual(my_obj.vice, None)
        my_obj.vice = "miami"
        self.assertEqual(my_obj.vice, "miami")

    def test_attributes_can_replace_internals(self):
        my_obj = fudge.Fake().has_attr(provides="hijacked")
        self.assertEqual(my_obj.provides, "hijacked")

    def test_repr_shortens_long_values(self):
        fake = (
            Fake("widget")
            .provides("set_bits")
            .with_args("12345678910111213141516171819202122232425262728293031")
        )
        try:
            fake.set_bits()
        except AssertionError as exc:
            self.assertEqual(
                str(exc),
                "fake:widget.set_bits('123456789101112131415161718192021222324252627...') "
                "was called unexpectedly with args ()",
            )
        else:
            raise RuntimeError("expected AssertionError")


class TestChainedNames(unittest.TestCase):

    def setUp(self):
        self.fake = fudge.Fake("db.Adapter")

    def tearDown(self):
        fudge.clear_expectations()

    def test_basic(self):
        self.assertEqual(repr(self.fake), "fake:db.Adapter")

    def test_nesting(self):
        f = self.fake
        f = f.provides("query").returns_fake().provides("fetchall")
        self.assertEqual(repr(f), "fake:db.Adapter.query()")
        f = f.provides("cursor").returns_fake()
        self.assertEqual(repr(f), "fake:db.Adapter.query().cursor()")

    def test_more_nesting(self):
        class ctx:
            fake = None

        @fudge.patch("smtplib.SMTP")
        def test(fake_SMTP):
            (fake_SMTP.is_callable().returns_fake().provides("sendmail"))
            ctx.fake = fake_SMTP()

        test()
        self.assertEqual(str(ctx.fake), "fake:smtplib.SMTP()")


class TestIsAStub(unittest.TestCase):

    def tearDown(self):
        fudge.clear_expectations()

    def test_is_callable(self):
        f = fudge.Fake().is_a_stub()
        result = f()
        assert isinstance(result, fudge.Fake)

    def test_infinite_callables(self):
        f = fudge.Fake().is_a_stub()
        result = f()()()
        assert isinstance(result, fudge.Fake)

    def test_is_any_call(self):
        f = fudge.Fake().is_a_stub()
        assert isinstance(f.foobar(), fudge.Fake)
        assert isinstance(f.foozilated(), fudge.Fake)

    def test_is_any_call_with_any_args(self):
        f = fudge.Fake().is_a_stub()
        assert isinstance(f.foobar(blazz="Blaz", kudoz="Klazzed"), fudge.Fake)

    def test_stubs_are_infinite(self):
        f = fudge.Fake().is_a_stub()
        assert isinstance(f.one().two().three(), fudge.Fake)

    def test_stubs_have_any_attribute(self):
        f = fudge.Fake().is_a_stub()
        assert isinstance(f.foo, fudge.Fake)
        assert isinstance(f.bar, fudge.Fake)

    def test_attributes_are_infinite(self):
        f = fudge.Fake().is_a_stub()
        assert isinstance(f.foo.bar.barfoo, fudge.Fake)

    def test_infinite_path_expectation(self):
        f = fudge.Fake().is_a_stub()
        f.foo.bar().expects("barfoo")
        f.foo.bar().barfoo()

    def test_infinite_path_expectation_is_verified(self):
        with self.assertRaises(AssertionError):
            f = fudge.Fake().is_a_stub()
            f.foo.bar().expects("barfoo").with_args(foo="bar")
            f.foo.bar().barfoo()
            fudge.verify()

    def test_infinite_path_naming(self):
        f = fudge.Fake(name="base").is_a_stub()
        self.assertEqual(str(f.foo.bar().baz), "fake:base.foo.bar().baz")


class TestLongArgValues(unittest.TestCase):

    def test_arg_diffs_are_not_shortened(self):
        fake = (
            Fake("widget")
            .provides("set_bits")
            .with_args("12345678910111213141516171819202122232425262728293031")
        )
        try:
            # this should not be shortened but the above arg spec should:
            fake.set_bits("99999999999999999999999999999999999999999999999999999999")
        except AssertionError as exc:
            self.assertEqual(
                str(exc),
                "fake:widget.set_bits('123456789101112131415161718192021222324252627...') "
                "was called unexpectedly with args "
                "('99999999999999999999999999999999999999999999999999999999')",
            )
        else:
            raise RuntimeError("expected AssertionError")

    def test_kwarg_diffs_are_not_shortened(self):
        fake = (
            Fake("widget")
            .provides("set_bits")
            .with_args(newbits="12345678910111213141516171819202122232425262728293031")
        )
        try:
            # this should not be shortened but the above arg spec should:
            fake.set_bits(newbits="99999999999999999999999999999999999999999999999999999999")
        except AssertionError as exc:
            self.assertEqual(
                str(exc),
                "fake:widget.set_bits(newbits='123456789101112131415161718192021222324252627...') "
                "was called unexpectedly with args "
                "(newbits='99999999999999999999999999999999999999999999999999999999')",
            )
        else:
            raise RuntimeError("expected AssertionError")


class TestArguments(unittest.TestCase):

    def setUp(self):
        self.fake = fudge.Fake()

    def tearDown(self):
        fudge.clear_expectations()

    def test_wrong_args(self):
        with self.assertRaises(AssertionError):
            exp = self.fake.expects("theCall").with_args(
                1, "ho ho ho ho ho ho ho, it's Santa", {"ditto": False}
            )
            exp.theCall()

    def test_wrong_kwargs(self):
        with self.assertRaises(AssertionError):
            exp = self.fake.expects("other").with_args(one="twozy", items=[1, 2, 3, 4])
            exp.other(nice="NOT NICE")

    def test_arg_count(self):
        with self.assertRaises(AssertionError):
            exp = self.fake.expects("one").with_arg_count(3)
            exp.one("no", "maybe")

    def test_kwarg_count(self):
        with self.assertRaises(AssertionError):
            exp = self.fake.expects("__init__").with_kwarg_count(2)
            exp(maybe="yes, maybe")

    def test_with_args_requires_a_method(self):
        with self.assertRaises(FakeDeclarationError):
            self.fake.with_args("something")

    def test_with_args_can_operate_on_provision(self):
        with self.assertRaises(AssertionError):
            self.fake.provides("not_expected").with_args("something")
            self.fake.not_expected()  # should still raise arg error

    def test_with_args_checks_args(self):
        with self.assertRaises(AssertionError):
            self.fake.expects("count").with_args("one", two="two")
            self.fake.count(two="two")

    def test_with_args_checks_kwargs(self):
        with self.assertRaises(AssertionError):
            self.fake.expects("count").with_args("one", two="two")
            self.fake.count("one")

    def test_raises_does_not_obscure_with_kwargs(self):
        with self.assertRaises(AssertionError):
            # previously, this test failed because the raises(exc)
            # was raised too early.  Issue 6
            self.fake.expects("count").with_args(two="two").raises(RuntimeError("bork"))
            self.fake.count("one")  # wrong kwargs

    def test_raises_does_not_obscure_with_args(self):
        with self.assertRaises(AssertionError):
            # Issue 6
            self.fake.expects("count").with_args("one").raises(RuntimeError("bork"))
            self.fake.count(two="two")  # wrong args

    def test_too_many_args(self):
        with self.assertRaises(AssertionError):
            db = Fake("db").expects("execute").with_args(bind={"one": 1})
            db.execute("select foozilate()", bind={"one": 1})  # unexpected statement arg

    def test_zero_keywords_ok(self):
        mail = fudge.Fake("mail").expects("send").with_arg_count(3)
        mail.send("you", "me", "hi")  # no kw args should be ok

    def test_zero_args_ok(self):
        mail = fudge.Fake("mail").expects("send").with_kwarg_count(3)
        mail.send(to="you", from_="me", msg="hi")  # no args should be ok

    def test_with_args_with_object_that_is_never_equal_to_anything(self):
        class NeverEqual(object):
            def __eq__(self, other):
                return False

        obj = NeverEqual()
        self.fake.expects("save").with_args(arg.any())
        self.fake.save(obj)  # this should pass but was failing prior to issue 9

    def test_with_kwargs_with_object_that_is_never_equal_to_anything(self):
        class NeverEqual(object):
            def __eq__(self, other):
                return False

        obj = NeverEqual()
        self.fake.expects("save").with_args(foo=arg.any())
        self.fake.save(foo=obj)  # this should pass but was failing prior to issue 9

    def test_with_matching_positional_args(self):
        db = self.fake
        db.expects("transaction").with_matching_args("insert")
        db.transaction("insert", isolation_level="lock")
        fudge.verify()

    def test_with_matching_keyword_args(self):
        db = self.fake
        db.expects("transaction").with_matching_args(isolation_level="lock")
        db.transaction("insert", isolation_level="lock")
        fudge.verify()

    def test_with_non_matching_positional_args(self):
        with self.assertRaises(AssertionError):
            db = self.fake
            db.expects("transaction").with_matching_args("update")
            db.transaction("insert", isolation_level="lock")
            fudge.verify()

    def test_with_too_many_non_matching_positional_args(self):
        with self.assertRaises(AssertionError):
            # this may be unintuitve but specifying too many
            # arguments constitutes as non-matching.  Why?
            # Because how else is it possible to implement, by index?
            db = self.fake
            db.expects("transaction").with_matching_args("update")
            db.transaction("update", "delete", isolation_level="lock")
            fudge.verify()

    def test_with_non_matching_keyword_args(self):
        with self.assertRaises(AssertionError):
            db = self.fake
            db.expects("transaction").with_matching_args(isolation_level="read")
            db.transaction("insert", isolation_level="lock")
            fudge.verify()

    def test_missing_matching_positional_args_is_not_ok(self):
        with self.assertRaises(AssertionError):
            # this is awkward to implement so I think it should not be supported
            db = self.fake
            db.expects("transaction").with_matching_args("update")
            db.transaction()
            fudge.verify()

    def test_missing_matching_keyword_args_is_ok(self):
        db = self.fake
        db.expects("transaction").with_matching_args(isolation_level="read")
        db.transaction()
        fudge.verify()


class TestCall(unittest.TestCase):

    def setUp(self):
        self.fake = fudge.Fake("SMTP")

    def test_repr(self):
        s = Call(self.fake)
        self.assertEqual(repr(s), "fake:SMTP()")

    def test_repr_callable(self):
        s = Call(self.fake.is_callable())
        self.assertEqual(repr(s), "fake:SMTP()")

    def test_repr_with_args(self):
        s = Call(self.fake)
        s.expected_args = [1, "bad"]
        self.assertEqual(repr(s), "fake:SMTP(1, 'bad')")

    def test_repr_with_kwargs(self):
        s = Call(self.fake)
        s.expected_args = [1, "bad"]
        s.expected_kwargs = {"baz": "borzo"}
        self.assertEqual(repr(s), "fake:SMTP(1, 'bad', baz='borzo')")

    def test_named_repr_with_args(self):
        s = Call(self.fake, call_name="connect")
        s.expected_args = [1, "bad"]
        self.assertEqual(repr(s), "fake:SMTP.connect(1, 'bad')")

    def test_nested_named_repr_with_args(self):
        f = self.fake.provides("get_conn").returns_fake()
        s = Call(f, call_name="connect")
        s.expected_args = [1, "bad"]
        self.assertEqual(repr(s), "fake:SMTP.get_conn().connect(1, 'bad')")

    def test_named_repr_with_index(self):
        s = Call(self.fake, call_name="connect")
        s.expected_args = [1, "bad"]
        s.index = 0
        self.assertEqual(repr(s), "fake:SMTP.connect(1, 'bad')[0]")
        s.index = 1
        self.assertEqual(repr(s), "fake:SMTP.connect(1, 'bad')[1]")


class TestCallStack(unittest.TestCase):

    def setUp(self):
        self.fake = fudge.Fake("SMTP")

    def test_calls(self):
        call_stack = CallStack(self.fake)

        c = Call(self.fake)
        c.return_val = 1
        call_stack.add_call(c)

        c = Call(self.fake)
        c.return_val = 2
        call_stack.add_call(c)

        self.assertEqual(call_stack(), 1)
        self.assertEqual(call_stack(), 2)

    def test_no_calls(self):
        with self.assertRaises(AssertionError):
            call_stack = CallStack(self.fake)
            call_stack()

    def test_end_of_calls(self):
        with self.assertRaises(AssertionError):
            call_stack = CallStack(self.fake)

            c = Call(self.fake)
            c.return_val = 1
            call_stack.add_call(c)

            self.assertEqual(call_stack(), 1)
            call_stack()

    def test_get_call_object(self):
        call_stack = CallStack(self.fake)

        c = Call(self.fake)
        call_stack.add_call(c)

        assert call_stack.get_call_object() is c

        d = Call(self.fake)
        call_stack.add_call(d)

        assert call_stack.get_call_object() is d

    def test_with_initial_calls(self):
        c = Call(self.fake)
        c.return_val = 1
        call_stack = CallStack(self.fake, initial_calls=[c])

        self.assertEqual(call_stack(), 1)

    def test_reset(self):
        call_stack = CallStack(self.fake)

        c = Call(self.fake)
        c.return_val = 1
        call_stack.add_call(c)

        c = Call(self.fake)
        c.return_val = 2
        call_stack.add_call(c)

        self.assertEqual(call_stack(), 1)
        self.assertEqual(call_stack(), 2)

        call_stack.reset()

        self.assertEqual(call_stack(), 1)
        self.assertEqual(call_stack(), 2)


class TestFakeCallables(unittest.TestCase):

    def tearDown(self):
        fudge.clear_expectations()

    def test_not_callable_by_default(self):
        with self.assertRaises(RuntimeError):
            self.fake = fudge.Fake()
            self.fake()

    def test_callable(self):
        fake = fudge.Fake().is_callable()
        fake()  # allow the call
        fudge.verify()  # no error

    def test_cannot_stub_any_call_by_default(self):
        with self.assertRaises(AttributeError):
            self.fake.Anything()  # must define this first

    def test_stub_with_args(self):
        with self.assertRaises(AssertionError):
            self.fake = fudge.Fake().is_callable().with_args(1, 2)
            self.fake(1)

    def test_stub_with_arg_count(self):
        with self.assertRaises(AssertionError):
            self.fake = fudge.Fake().is_callable().with_arg_count(3)
            self.fake("bah")

    def test_stub_with_kwarg_count(self):
        with self.assertRaises(AssertionError):
            self.fake = fudge.Fake().is_callable().with_kwarg_count(3)
            self.fake(two=1)

    def test_stub_with_provides(self):
        self.fake = fudge.Fake().provides("something")
        self.fake.something()

    def test_fake_can_sabotage_itself(self):
        # I'm not sure if there should be a warning
        # for this but it seems that it should be
        # possible for maximum flexibility:

        # replace Fake.with_args()
        self.fake = fudge.Fake().provides("with_args").returns(1)
        self.assertEqual(self.fake.with_args(), 1)

    def test_stub_with_provides_and_args(self):
        with self.assertRaises(AssertionError):
            self.fake = fudge.Fake().provides("something").with_args(1, 2)
            self.fake.something()

    def test_stub_is_not_registered(self):
        self.fake = fudge.Fake().provides("something")
        exp = self.fake._get_current_call()
        self.assertEqual(exp.call_name, "something")
        assert exp not in fudge.registry

    def test_raises_class(self):
        with self.assertRaises(RuntimeError):
            self.fake = fudge.Fake().provides("fail").raises(RuntimeError)
            self.fake.fail()

    def test_raises_instance(self):
        with self.assertRaises(RuntimeError):
            self.fake = fudge.Fake().provides("fail").raises(RuntimeError("batteries ran out"))
            self.fake.fail()


class TestReplacementCalls(unittest.TestCase):

    def tearDown(self):
        fudge.clear_expectations()

    def test_replace_call(self):

        def something():
            return "hijacked"

        fake = fudge.Fake().provides("something").calls(something)
        self.assertEqual(fake.something(), "hijacked")

    def test_calls_mixed_with_returns(self):

        called = []

        def something():
            called.append(True)
            return "hijacked"

        fake = fudge.Fake().provides("something").calls(something).returns("other")
        self.assertEqual(fake.something(), "other")
        self.assertEqual(called, [True])

    def test_calls_mixed_with_expectations(self):
        def something():
            return "hijacked"

        with self.assertRaises(AssertionError):
            # with_args() expectation should not get lost:
            fake = fudge.Fake().provides("something").calls(something).with_args(1, 2)
            self.assertEqual(fake.something(), "hijacked")

    def test_replace_init(self):

        class custom_object:
            def hello(self):
                return "hi"

        fake = fudge.Fake().provides("__init__").returns(custom_object())
        self.assertEqual(fake().hello(), "hi")


class TestFakeTimesCalled(unittest.TestCase):

    def tearDown(self):
        fudge.clear_expectations()

    def test_when_provided(self):
        self.fake = fudge.Fake().provides("something").times_called(2)
        # this should not raise an error because the call was provided not expected
        fudge.verify()

    def test_when_provided_raises_on_too_many_calls(self):
        with self.assertRaises(AssertionError):
            self.fake = fudge.Fake().provides("something").times_called(2)
            self.fake.something()
            self.fake.something()
            self.fake.something()  # too many

    def test_when_expected(self):
        with self.assertRaises(AssertionError):
            self.fake = fudge.Fake().expects("something").times_called(2)
            self.fake.something()
            fudge.verify()

    def test_when_expected_raises_on_too_many_calls(self):
        with self.assertRaises(AssertionError):
            self.fake = fudge.Fake().expects("something").times_called(2)
            self.fake.something()
            self.fake.something()
            self.fake.something()  # too many
            fudge.verify()

    def test_expected_callable(self):
        with self.assertRaises(AssertionError):
            login = fudge.Fake("login", expect_call=True).times_called(2)
            login()
            fudge.verify()

    def test_callable_ok(self):
        self.fake = fudge.Fake(callable=True).times_called(2)
        self.fake()
        self.fake()
        fudge.verify()

    def test_when_provided_ok(self):
        self.fake = fudge.Fake().provides("something").times_called(2)
        self.fake.something()
        self.fake.something()
        fudge.verify()

    def test_when_expected_ok(self):
        self.fake = fudge.Fake().expects("something").times_called(2)
        self.fake.something()
        self.fake.something()
        fudge.verify()


class TestNextCall(unittest.TestCase):

    def tearDown(self):
        fudge.clear_expectations()

    def test_next_call_then_times_called_is_error(self):
        with self.assertRaises(FakeDeclarationError):
            self.fake = fudge.Fake().expects("hi").returns("goodday").next_call().times_called(4)
            self.fake.hi()
            self.fake.hi()
            fudge.verify()

    def test_times_called_then_next_call_is_error(self):
        with self.assertRaises(FakeDeclarationError):
            self.fake = fudge.Fake().expects("hi").times_called(4).next_call()
            self.fake.hi()
            self.fake.hi()
            fudge.verify()

    def test_stacked_returns(self):
        fake = fudge.Fake().provides("something")
        fake = fake.returns(1)
        fake = fake.next_call()
        fake = fake.returns(2)
        fake = fake.next_call()
        fake = fake.returns(3)

        self.assertEqual(fake.something(), 1)
        self.assertEqual(fake.something(), 2)
        self.assertEqual(fake.something(), 3)

    def test_stacked_calls_are_finite(self):
        with self.assertRaises(AssertionError):
            self.fake = fudge.Fake().provides("something")
            self.fake = self.fake.returns(1)
            self.fake = self.fake.next_call()
            self.fake = self.fake.returns(2)

            self.assertEqual(self.fake.something(), 1)
            self.assertEqual(self.fake.something(), 2)
            self.fake.something()

    def test_stack_is_reset_when_name_changes(self):
        self.fake = fudge.Fake().provides("something")
        self.fake = self.fake.returns(1)
        self.fake = self.fake.next_call()
        self.fake = self.fake.returns(2)
        self.fake = self.fake.provides("other")
        self.fake = self.fake.returns(3)

        self.assertEqual(self.fake.something(), 1)
        self.assertEqual(self.fake.something(), 2)
        self.assertEqual(self.fake.other(), 3)
        self.assertEqual(self.fake.other(), 3)
        self.assertEqual(self.fake.other(), 3)
        self.assertEqual(self.fake.other(), 3)
        self.assertEqual(self.fake.other(), 3)

    def test_next_call_with_multiple_returns(self):
        self.fake = fudge.Fake().provides("something")
        self.fake = self.fake.returns(1)
        self.fake = self.fake.next_call()
        self.fake = self.fake.returns(2)
        self.fake = self.fake.provides("other")
        self.fake = self.fake.returns(3)
        self.fake = self.fake.next_call()
        self.fake = self.fake.returns(4)

        self.assertEqual(self.fake.something(), 1)
        self.assertEqual(self.fake.something(), 2)
        self.assertEqual(self.fake.other(), 3)
        self.assertEqual(self.fake.other(), 4)

    def test_stacked_calls_do_not_collide(self):
        self.fake = fudge.Fake().provides("something")
        self.fake = self.fake.returns(1)
        self.fake = self.fake.next_call()
        self.fake = self.fake.returns(2)
        self.fake = self.fake.provides("other")
        self.fake = self.fake.returns(3)
        self.fake = self.fake.next_call()
        self.fake = self.fake.returns(4)

        self.assertEqual(self.fake.something(), 1)
        self.assertEqual(self.fake.other(), 3)
        self.assertEqual(self.fake.something(), 2)
        self.assertEqual(self.fake.other(), 4)

    def test_returns_are_infinite(self):
        self.fake = fudge.Fake().provides("something").returns(1)

        self.assertEqual(self.fake.something(), 1)
        self.assertEqual(self.fake.something(), 1)
        self.assertEqual(self.fake.something(), 1)
        self.assertEqual(self.fake.something(), 1)
        self.assertEqual(self.fake.something(), 1)
        self.assertEqual(self.fake.something(), 1)
        self.assertEqual(self.fake.something(), 1)
        self.assertEqual(self.fake.something(), 1)

    def test_stacked_does_not_copy_expectations(self):

        fake = fudge.Fake().expects("add")
        fake = fake.with_args(1, 2).returns(3)

        fake = fake.next_call()
        fake = fake.returns(-1)

        self.assertEqual(fake.add(1, 2), 3)
        self.assertEqual(fake.add(), -1)

    def test_stacked_calls_are_in_registry(self):
        fake = fudge.Fake().expects("count").with_args(1)
        fake = fake.next_call().with_args(2)
        fake = fake.next_call().with_args(3)
        fake = fake.next_call().with_args(4)

        # hmm
        call_stack = fake._declared_calls[fake._last_declared_call_name]
        calls = [c for c in call_stack]
        assert calls[0] in fudge.registry
        assert calls[1] in fudge.registry
        assert calls[2] in fudge.registry
        assert calls[3] in fudge.registry

    def test_stacked_calls_are_indexed(self):
        fake = fudge.Fake().expects("count").with_args(1)
        fake = fake.next_call().with_args(2)
        fake = fake.next_call().with_args(3)
        fake = fake.next_call().with_args(4)

        # hmm
        call_stack = fake._declared_calls[fake._last_declared_call_name]
        calls = [c for c in call_stack]
        self.assertEqual(calls[0].index, 0)
        self.assertEqual(calls[1].index, 1)
        self.assertEqual(calls[2].index, 2)
        self.assertEqual(calls[3].index, 3)

    def test_start_stop_resets_stack(self):
        fudge.clear_expectations()
        fake = fudge.Fake().provides("something")
        fake = fake.returns(1)
        fake = fake.next_call()
        fake = fake.returns(2)

        self.assertEqual(fake.something(), 1)
        self.assertEqual(fake.something(), 2)

        fudge.clear_calls()

        self.assertEqual(fake.something(), 1)
        self.assertEqual(fake.something(), 2)

        fudge.verify()

        self.assertEqual(fake.something(), 1)
        self.assertEqual(fake.something(), 2)

    def test_next_call_with_callables(self):
        login = (
            fudge.Fake("login")
            .is_callable()
            .returns("yes")
            .next_call()
            .returns("maybe")
            .next_call()
            .returns("no")
        )
        self.assertEqual(login(), "yes")
        self.assertEqual(login(), "maybe")
        self.assertEqual(login(), "no")

    def test_returns(self):
        db = (
            Fake("db")
            .provides("get_id")
            .returns(1)
            .provides("set_id")
            .next_call(for_method="get_id")
            .returns(2)
        )
        # print [c.return_val for c in db._declared_calls["get_id"]._calls]
        self.assertEqual(db.get_id(), 1)
        self.assertEqual(db.set_id(), None)
        self.assertEqual(db.get_id(), 2)

    def test_expectations_with_multiple_return_values(self):
        db = (
            Fake("db")
            .expects("get_id")
            .returns(1)
            .expects("set_id")
            .next_call(for_method="get_id")
            .returns(2)
        )
        self.assertEqual(db.get_id(), 1)
        self.assertEqual(db.set_id(), None)
        self.assertEqual(db.get_id(), 2)

        fudge.verify()


class TestExpectsAndProvides(unittest.TestCase):

    def tearDown(self):
        fudge.clear_expectations()

    def test_nocall(self):
        with self.assertRaises(AssertionError):
            fake = fudge.Fake()
            fake.expects("something")
            fudge.verify()

    def test_multiple_provides_on_chained_fakes_ok(self):
        Fake("db").provides("insert").returns_fake().provides("insert")

    def test_multiple_expects_on_chained_fakes_ok(self):
        Fake("db").expects("insert").returns_fake().expects("insert")

    def test_callable_expectation(self):
        fake_setup = fudge.Fake("setup", expect_call=True)
        fake_setup()
        # was called so verification should pass:
        fudge.verify()

    def test_callable_expectation_with_args(self):
        fake_setup = fudge.Fake("setup", expect_call=True).with_args("<db>")
        fake_setup("<db>")
        # was called so verification should pass:
        fudge.verify()

    def test_multiple_expects_act_like_next_call(self):
        self.fake = fudge.Fake().expects("something")
        self.fake = self.fake.returns(1)
        self.fake = self.fake.expects("something")
        self.fake = self.fake.returns(2)

        self.assertEqual(self.fake.something(), 1)
        self.assertEqual(self.fake.something(), 2)

    def test_multiple_provides_act_like_next_call(self):
        self.fake = fudge.Fake().provides("something")
        self.fake = self.fake.returns(1)
        self.fake = self.fake.provides("something")
        self.fake = self.fake.returns(2)

        self.assertEqual(self.fake.something(), 1)
        self.assertEqual(self.fake.something(), 2)

    def test_multiple_expects_for_sep_methods(self):
        self.fake = (
            fudge.Fake()
            .expects("marco")
            .returns(1)
            .expects("polo")
            .returns("A")
            .expects("marco")
            .returns(2)
            .expects("polo")
            .returns("B")
        )

        self.assertEqual(self.fake.marco(), 1)
        self.assertEqual(self.fake.marco(), 2)
        self.assertEqual(self.fake.polo(), "A")
        self.assertEqual(self.fake.polo(), "B")

    def test_multiple_provides_for_sep_methods(self):
        self.fake = (
            fudge.Fake()
            .provides("marco")
            .returns(1)
            .provides("polo")
            .returns("A")
            .provides("marco")
            .returns(2)
            .provides("polo")
            .returns("B")
        )

        self.assertEqual(self.fake.marco(), 1)
        self.assertEqual(self.fake.marco(), 2)
        self.assertEqual(self.fake.polo(), "A")
        self.assertEqual(self.fake.polo(), "B")


class TestOrderedCalls(unittest.TestCase):

    def tearDown(self):
        fudge.clear_expectations()

    def test_out_of_order(self):
        with self.assertRaises(AssertionError):
            fake = fudge.Fake().remember_order().expects("one").expects("two")
            fake.two()
            fake.one()
            fudge.verify()

    def test_cannot_remember_order_when_callable_is_true(self):
        with self.assertRaises(FakeDeclarationError):
            fudge.Fake().is_callable().remember_order()

    def test_cannot_remember_order_when_expect_call_is_true(self):
        with self.assertRaises(FakeDeclarationError):
            fudge.Fake(expect_call=True).remember_order()

    def test_not_enough_calls(self):
        with self.assertRaises(AssertionError):
            # need to drop down a level to bypass expected calls:
            r = Registry()
            fake = Fake()
            call_order = ExpectedCallOrder(fake)
            r.remember_expected_call_order(call_order)

            exp = ExpectedCall(fake, "callMe", call_order=call_order)
            call_order.add_expected_call(exp)

            r.verify()

    def test_only_one_call(self):
        with self.assertRaises(AssertionError):
            # need to drop down a level to bypass expected calls:
            r = Registry()
            fake = Fake()
            call_order = ExpectedCallOrder(fake)
            r.remember_expected_call_order(call_order)

            exp = ExpectedCall(fake, "one", call_order=call_order)
            call_order.add_expected_call(exp)
            exp()  # call this

            exp = ExpectedCall(fake, "two", call_order=call_order)
            call_order.add_expected_call(exp)

            r.verify()

    def test_incremental_order_assertion_ok(self):
        # need to drop down a level to bypass expected calls:
        fake = Fake()
        call_order = ExpectedCallOrder(fake)

        exp = ExpectedCall(fake, "one", call_order=call_order)
        call_order.add_expected_call(exp)
        exp()  # call this

        exp = ExpectedCall(fake, "two", call_order=call_order)
        call_order.add_expected_call(exp)

        # two() not called but assertion is not finalized:
        call_order.assert_order_met(finalize=False)

    def test_multiple_returns_affect_order(self):
        db = (
            Fake("db")
            .remember_order()
            .expects("get_id")
            .returns(1)
            .expects("set_id")
            .next_call(for_method="get_id")
            .returns(2)
        )
        self.assertEqual(db.get_id(), 1)
        self.assertEqual(db.set_id(), None)
        self.assertEqual(db.get_id(), 2)
        fudge.verify()

    def test_chained_fakes_honor_order(self):
        with self.assertRaises(AssertionError):
            Thing = Fake("thing").remember_order().expects("__init__")
            holder = Thing.expects("get_holder").returns_fake()
            holder = holder.expects("init")

            thing = Thing()
            holder = thing.get_holder()
            # missing thing.init()
            fudge.verify()

    def test_too_many_calls(self):
        with self.assertRaises(AssertionError):
            db = Fake("db").remember_order().expects("get_id").returns(1).expects("set_id")
            self.assertEqual(db.get_id(), 1)
            self.assertEqual(db.set_id(), None)
            # extra :
            self.assertEqual(db.get_id(), 1)

    def test_expects_call_shortcut(self):
        with self.assertRaises(AssertionError):
            remove = Fake("os.remove").expects_call()
            fudge.verify()
            assert isinstance(remove, Fake)

    def test_expects_call_shortcut_ok(self):
        remove = Fake("os.remove").expects_call()
        remove()
        fudge.verify()
        assert isinstance(remove, Fake)

    def test_provides_call_shortcut(self):
        remove = Fake("os.remove").is_callable()
        remove()
        assert isinstance(remove, Fake)


class TestPatchedFakes(unittest.TestCase):

    def tearDown(self):
        fudge.clear_expectations()

    def test_expectations_are_cleared(self):

        class holder:
            test_called = False

        # Set up decoy expectation:
        fudge.Fake("db").expects("save")

        @fudge.patch("shutil.copy")
        def some_test(copy):
            holder.test_called = True

        some_test()
        fudge.verify()  # should be no errors
        self.assertEqual(holder.test_called, True)

    def test_expectations_are_verified(self):

        class holder:
            test_called = False

        @fudge.patch("shutil.copy")
        def some_test(copy):
            copy.expects("__call__")
            holder.test_called = True

        with self.assertRaises(AssertionError):
            some_test()
        self.assertEqual(holder.test_called, True)

    def test_expectations_are_always_cleared(self):

        class holder:
            test_called = False

        @fudge.patch("shutil.copy")
        def some_test(copy):
            holder.test_called = True
            copy.expects_call()
            raise RuntimeError

        with self.assertRaises(RuntimeError):
            some_test()
        fudge.verify()  # should be no errors
        self.assertEqual(holder.test_called, True)

    def test_calls_are_cleared(self):

        class holder:
            test_called = False

        sess = fudge.Fake("session").expects("save")
        # call should be cleared:
        sess.save()

        @fudge.patch("shutil.copy")
        def some_test(copy):
            holder.test_called = True
            copy.expects_call().times_called(1)
            copy()

        some_test()
        fudge.verify()  # should be no errors
        self.assertEqual(holder.test_called, True)

    def test_with_statement(self):

        class holder:
            test_called = False

        def run_test():
            with self.assertRaises(AssertionError):
                with fudge.patch("shutil.copy") as copy:
                    copy.expects("__call__")
                    holder.test_called = True

        run_test()
        self.assertEqual(holder.test_called, True)

    def test_with_statement_exception(self):

        def run_test():
            with self.assertRaises(RuntimeError):
                with fudge.patch("shutil.copy") as copy:
                    copy.expects("__call__")
                    raise RuntimeError()

        run_test()


class TestNonPatchedFakeTest(unittest.TestCase):

    def tearDown(self):
        fudge.clear_expectations()

    def test_preserve_method(self):
        class holder:
            test_called = False

        @fudge.test
        def some_test():
            holder.test_called = True

        self.assertEqual(some_test.__name__, "some_test")

    def test_expectations_are_cleared(self):

        class holder:
            test_called = False

        # Set up decoy expectation:
        fudge.Fake("db").expects("save")

        @fudge.test
        def some_test():
            holder.test_called = True

        some_test()
        fudge.verify()  # should be no errors
        self.assertEqual(holder.test_called, True)

    def test_expectations_are_always_cleared(self):

        class holder:
            test_called = False

        @fudge.test
        def some_test():
            holder.test_called = True
            fudge.Fake("db").expects("save")
            raise RuntimeError

        with self.assertRaises(RuntimeError):
            some_test()
        fudge.verify()  # should be no errors
        self.assertEqual(holder.test_called, True)

    def test_calls_are_cleared(self):

        class holder:
            test_called = False

        sess = fudge.Fake("session").expects("save")
        # call should be cleared:
        sess.save()

        @fudge.test
        def some_test():
            holder.test_called = True
            db = fudge.Fake("db").expects("save").times_called(1)
            db.save()

        some_test()
        fudge.verify()  # should be no errors
        self.assertEqual(holder.test_called, True)

    def test_verify(self):

        @fudge.test
        def some_test():
            fudge.Fake("db").expects("save")

        with self.assertRaises(AssertionError):
            some_test()



"""
Fudge replaces real objects with fakes (mocks, stubs) for testing purposes.
"""

__version__ = "1.1.1"

from functools import wraps
import os
import re
import sys
import threading
import warnings
from fudge.exc import FakeDeclarationError
from fudge.patcher import (  # noqa: F401
    patch,
    with_patched_object,
    patched_context,
    patch_object,
    NonExistant,
    PatchHandler,
)
from fudge.util import fmt_val, fmt_dict_vals

__all__ = ["Fake", "patch", "test", "clear_calls", "verify", "clear_expectations"]


class Registry(object):
    """An internal, thread-safe registry of expected calls.

    You do not need to use this directly, use Fake.expects(...), etc
    """

    def __init__(self):
        self.expected_calls = {}
        self.expected_call_order = {}
        self.call_stacks = []

    def __contains__(self, obj):
        return obj in self.get_expected_calls()

    def clear_actual_calls(self):
        for exp in self.get_expected_calls():
            exp.was_called = False

    def clear_all(self):
        self.clear_actual_calls()
        self.clear_expectations()

    def clear_calls(self):
        """Clears out any calls that were made on previously
        registered fake objects and resets all call stacks.

        You do not need to use this directly.  Use fudge.clear_calls()
        """
        self.clear_actual_calls()
        for stack in self.call_stacks:
            stack.reset()
        for fake, call_order in self.get_expected_call_order().items():
            call_order.reset_calls()

    def clear_expectations(self):
        c = self.get_expected_calls()
        c[:] = []
        d = self.get_expected_call_order()
        d.clear()

    def expect_call(self, expected_call):
        c = self.get_expected_calls()
        c.append(expected_call)
        call_order = self.get_expected_call_order()
        if expected_call.fake in call_order:
            this_call_order = call_order[expected_call.fake]
            this_call_order.add_expected_call(expected_call)

    def get_expected_calls(self):
        self.expected_calls.setdefault(threading.get_ident(), [])
        return self.expected_calls[threading.get_ident()]

    def get_expected_call_order(self):
        self.expected_call_order.setdefault(threading.get_ident(), {})
        return self.expected_call_order[threading.get_ident()]

    def remember_expected_call_order(self, expected_call_order):
        ordered_fakes = self.get_expected_call_order()
        fake = expected_call_order.fake
        # does nothing if called twice like:
        # Fake().remember_order().remember_order()
        ordered_fakes.setdefault(fake, expected_call_order)

    def register_call_stack(self, call_stack):
        self.call_stacks.append(call_stack)

    def verify(self):
        """Ensure all expected calls were called,
        raise AssertionError otherwise.

        You do not need to use this directly.  Use fudge.verify()
        """
        try:
            for exp in self.get_expected_calls():
                exp.assert_called()
                exp.assert_times_called()
            for fake, call_order in self.get_expected_call_order().items():
                call_order.assert_order_met(finalize=True)
        finally:
            self.clear_calls()


registry = Registry()


def clear_calls():
    """Begin a new set of calls on fake objects.

    Specifically, clear out any calls that
    were made on previously registered fake
    objects and reset all call stacks.
    You should call this any time you begin
    making calls on fake objects.

    This is also available in :func:`fudge.patch`, :func:`fudge.test` and :func:`fudge.with_fakes`
    """
    registry.clear_calls()


def verify():
    """Verify that all methods have been called as expected.

    Specifically, analyze all registered fake
    objects and raise an AssertionError if an
    expected call was never made to one or more
    objects.

    This is also available in :func:`fudge.patch`, :func:`fudge.test` and :func:`fudge.with_fakes`
    """
    registry.verify()


def start():
    """Start testing with fake objects.

    Deprecated.  Use :func:`fudge.clear_calls` instead.
    """
    warnings.warn(
        "fudge.start() has been deprecated.  Use fudge.clear_calls() instead", DeprecationWarning, 3
    )
    clear_calls()


def stop():
    """Stop testing with fake objects.

    Deprecated.  Use :func:`fudge.verify` instead.
    """
    warnings.warn(
        "fudge.stop() has been deprecated.  Use fudge.verify() instead", DeprecationWarning, 3
    )
    verify()


def clear_expectations():
    registry.clear_expectations()


def with_fakes(method):
    """
    Decorator that calls :func:`fudge.clear_calls` before method()
    and :func:`fudge.verify` afterwards.
    """

    @wraps(method)
    def apply_clear_and_verify(*args, **kw):
        clear_calls()
        method(*args, **kw)
        verify()  # if no exceptions

    return apply_clear_and_verify


def test(method):
    """Decorator for a test that uses fakes directly (not patched).

    Most of the time you probably want to use :func:`fudge.patch` instead.

    .. doctest::
        :hide:

        >>> import fudge

    .. doctest::

        >>> @fudge.test
        ... def test():
        ...     db = fudge.Fake('db').expects('connect')
        ...     # do stuff...
        ...
        >>> test()
        Traceback (most recent call last):
        ...
        AssertionError: fake:db.connect() was not called

    .. doctest::
        :hide:

        >>> fudge.clear_expectations()

    """

    @wraps(method)
    def clear_and_verify(*args, **kw):
        clear_expectations()
        clear_calls()
        try:
            v = method(*args, **kw)
            verify()  # if no exceptions
        finally:
            clear_expectations()
        return v

    return clear_and_verify


test.__test__ = False  # Nose: do not collect


class Call(object):
    """A call that can be made on a Fake object.

    You do not need to use this directly, use Fake.provides(...), etc

    index=None
        When numerical, this indicates the position of the call
        (as in, a CallStack)

    callable=False
        Means this object acts like a function, not a method of an
        object.

    call_order=ExpectedCallOrder()
        A call order to append each call to.  Default is None
    """

    def __init__(self, fake, call_name=None, index=None, callable=False, call_order=None):
        self.fake = fake
        self.call_name = call_name
        self.call_replacement = None
        self.expected_arg_count = None
        self.expected_kwarg_count = None
        self.expected_args = None
        self.expected_kwargs = None
        self.expected_matching_args = None
        self.expected_matching_kwargs = None
        self.unexpected_args = None
        self.unexpected_kwargs = None
        self.index = index
        self.exception_to_raise = None
        self.return_val = None
        self.was_called = False
        self.expected_times_called = None
        self.actual_times_called = 0
        self.callable = callable
        self.call_order = call_order

    def __call__(self, *args, **kwargs):
        self.was_called = True
        self.actual_times_called += 1
        if self.call_order:
            self.call_order.add_actual_call(self)
            self.call_order.assert_order_met(finalize=False)

        # make sure call count doesn't go over :
        if (
            self.expected_times_called is not None
            and self.actual_times_called > self.expected_times_called
        ):
            raise AssertionError(
                "%s was called %s time(s). Expected %s."
                % (self, self.actual_times_called, self.expected_times_called)
            )

        replacement_return = None

        if self.call_replacement:
            replacement_return = self.call_replacement(*args, **kwargs)
        if self.return_val is not None:
            # this wins:
            return_value = self.return_val
        else:
            # but it is intuitive to otherwise
            # return the replacement's return:
            return_value = replacement_return

        # determine whether we should inspect arguments or not:
        with_args = self.expected_args or self.expected_kwargs

        if with_args:
            # check keyword args first because of python arg coercion...
            if self.expected_kwargs is None:
                self.expected_kwargs = {}  # empty **kw
            if self.expected_kwargs != kwargs:
                raise AssertionError(
                    "%s was called unexpectedly with args %s"
                    % (self, self._repr_call(args, kwargs, shorten_long_vals=False))
                )

            if self.expected_args is None:
                self.expected_args = tuple([])  # empty *args
            if self.expected_args != args:
                raise AssertionError(
                    "%s was called unexpectedly with args %s"
                    % (self, self._repr_call(args, kwargs, shorten_long_vals=False))
                )

        # now check for matching keyword args.
        # i.e. keyword args that are only checked if the call provided them
        if self.expected_matching_kwargs:
            for expected_arg, expected_value in self.expected_matching_kwargs.items():
                if expected_arg in kwargs:
                    if expected_value != kwargs[expected_arg]:
                        raise AssertionError(
                            "%s was called unexpectedly with args %s"
                            % (
                                self,
                                self._repr_call(
                                    args,
                                    {expected_arg: kwargs[expected_arg]},
                                    shorten_long_vals=False,
                                ),
                            )
                        )

        # now check for matching args.
        # i.e. args that are only checked if the call provided them
        if self.expected_matching_args:
            if self.expected_matching_args != args:
                raise AssertionError(
                    "%s was called unexpectedly with args %s"
                    % (self, self._repr_call(args, kwargs, shorten_long_vals=False))
                )

        # determine whether we should inspect argument counts or not:
        with_arg_counts = (
            self.expected_arg_count is not None or self.expected_kwarg_count is not None
        )

        if with_arg_counts:
            if self.expected_arg_count is None:
                self.expected_arg_count = 0
            if len(args) != self.expected_arg_count:
                raise AssertionError(
                    "%s was called with %s arg(s) but expected %s"
                    % (self, len(args), self.expected_arg_count)
                )

            if self.expected_kwarg_count is None:
                self.expected_kwarg_count = 0
            if len(kwargs.keys()) != self.expected_kwarg_count:
                raise AssertionError(
                    "%s was called with %s keyword arg(s) but expected %s"
                    % (self, len(kwargs.keys()), self.expected_kwarg_count)
                )

        if self.unexpected_kwargs:
            for un_key, un_val in self.unexpected_kwargs.items():
                if un_key in kwargs and kwargs[un_key] == un_val:
                    raise AssertionError(
                        "%s was called unexpectedly with kwarg %s=%s" % (self, un_key, un_val)
                    )

        if self.unexpected_args:
            for un_arg in self.unexpected_args:
                if un_arg in args:
                    raise AssertionError("%s was called unexpectedly with arg %s" % (self, un_arg))

        if self.exception_to_raise is not None:
            raise self.exception_to_raise

        return return_value

    def _repr_call(self, expected_args, expected_kwargs, shorten_long_vals=True):
        args = []
        if expected_args:
            args.extend([fmt_val(a, shorten=shorten_long_vals) for a in expected_args])
        if expected_kwargs:
            args.extend(fmt_dict_vals(expected_kwargs, shorten=shorten_long_vals))
        if args:
            call = "(%s)" % ", ".join(args)
        else:
            call = "()"
        return call

    def __repr__(self):
        cls_name = repr(self.fake)
        if self.call_name and not self.callable:
            call = "%s.%s" % (cls_name, self.call_name)
        else:
            call = "%s" % cls_name
        call = "%s%s" % (call, self._repr_call(self.expected_args, self.expected_kwargs))
        if self.index is not None:
            call = "%s[%s]" % (call, self.index)
        return call

    def get_call_object(self):
        """return self.

        this exists for compatibility with :class:`CallStack`
        """
        return self

    def assert_times_called(self):
        if (
            self.expected_times_called is not None
            and self.actual_times_called != self.expected_times_called
        ):
            raise AssertionError(
                "%s was called %s time(s). Expected %s."
                % (self, self.actual_times_called, self.expected_times_called)
            )


class ExpectedCall(Call):
    """An expectation that a call will be made on a Fake object.

    You do not need to use this directly, use Fake.expects(...), etc
    """

    def __init__(self, *args, **kw):
        super(ExpectedCall, self).__init__(*args, **kw)
        registry.expect_call(self)

    def assert_called(self):
        if not self.was_called:
            raise AssertionError("%s was not called" % (self))


class ExpectedCallOrder(object):
    """An expectation that calls should be called in a specific order."""

    def __init__(self, fake):
        self.fake = fake
        self._call_order = []
        self._actual_calls = []

    def __repr__(self):
        return "%r(%r)" % (self.fake, self._call_order)

    __str__ = __repr__

    def _repr_call_list(self, call_list):
        if not len(call_list):
            return "no calls"
        else:
            stack = ["#%s %r" % (i + 1, c) for i, c in enumerate(call_list)]
            stack.append("end")
            return ", ".join(stack)

    def add_expected_call(self, call):
        self._call_order.append(call)

    def add_actual_call(self, call):
        self._actual_calls.append(call)

    def assert_order_met(self, finalize=False):
        """assert that calls have been made in the right order."""
        error = None
        actual_call_len = len(self._actual_calls)
        expected_call_len = len(self._call_order)

        if actual_call_len == 0:
            error = "Not enough calls were made"
        else:
            for i, call in enumerate(self._call_order):
                if actual_call_len < i + 1:
                    if not finalize:
                        # we're not done asserting calls so
                        # forget about not having enough calls
                        continue

                    calls_made = len(self._actual_calls)
                    if calls_made == 1:
                        error = "Only 1 call was made"
                    else:
                        error = "Only %s calls were made" % calls_made
                    break

                ac_call = self._actual_calls[i]
                if ac_call is not call:
                    error = "Call #%s was %r" % (i + 1, ac_call)
                    break

            if not error:
                if actual_call_len > expected_call_len:
                    # only show the first extra call since this
                    # will be triggered before all calls are finished:
                    error = "#%s %s was unexpected" % (
                        expected_call_len + 1,
                        self._actual_calls[expected_call_len],
                    )

        if error:
            msg = "%s; Expected: %s" % (error, self._repr_call_list(self._call_order))
            raise AssertionError(msg)

    def reset_calls(self):
        self._actual_calls[:] = []


class CallStack(object):
    """A stack of :class:`Call` objects

    Calling this object behaves just like Call except
    the Call instance you operate on gets changed each time __call__() is made

    expected=False
        When True, this indicates that the call stack was derived
        from an expected call.  This is used by Fake to register
        each call on the stack.

    call_name
        Name of the call

    """

    def __init__(self, fake, initial_calls=None, expected=False, call_name=None):
        self.fake = fake
        self._pointer = 0  # position of next call to be made (can be reset)
        self._calls = []
        if initial_calls is not None:
            for c in initial_calls:
                self.add_call(c)
        self.expected = expected
        self.call_name = call_name
        registry.register_call_stack(self)

    def __iter__(self):
        for c in self._calls:
            yield c

    def __repr__(self):
        return "<%s for %r>" % (self.__class__.__name__, self._calls)

    __str__ = __repr__

    def add_call(self, call):
        self._calls.append(call)
        call.index = len(self._calls) - 1

    def get_call_object(self):
        """returns the last *added* call object.

        this is so Fake knows which one to alter
        """
        return self._calls[len(self._calls) - 1]

    def reset(self):
        self._pointer = 0

    def __call__(self, *args, **kw):
        try:
            current_call = self._calls[self._pointer]
        except IndexError:
            raise AssertionError(
                "This attribute of %s can only be called %s time(s).  "
                "Call reset() if necessary or fudge.clear_calls()." % (self.fake, len(self._calls))
            )
        self._pointer += 1
        return current_call(*args, **kw)


class Fake(object):
    """A fake object that replaces a real one while testing.

    Most calls with a few exceptions return ``self`` so that you can chain
    them together to create readable code.

    Instance methods will raise either AssertionError or :class:`fudge.FakeDeclarationError`

    Keyword arguments:

    **name=None**
        Name of the class, module, or function you mean to replace. If not
        specified, Fake() will try to guess the name by inspecting the calling
        frame (if possible).

    **allows_any_call=False**
        This is **deprecated**.  Use :meth:`Fake:is_a_stub()` instead.

    **callable=False**
        This is **deprecated**.  Use :meth:`Fake.is_callable` instead.

    **expect_call=True**
        This is **deprecated**.  Use :meth:`Fake.expects_call` instead.

    """

    def __init__(self, name=None, allows_any_call=False, callable=False, expect_call=False):
        self._attributes = {}
        self._properties = {}
        self._declared_calls = {}
        self._name = name or self._guess_name()
        self._last_declared_call_name = None
        self._is_a_stub = False
        if allows_any_call:
            warnings.warn(
                "Fake(allows_any_call=True) is deprecated;" " use Fake.is_a_stub()",
                DeprecationWarning,
                3,
            )
            self.is_a_stub()
        self._call_stack = None
        if expect_call:
            self.expects_call()
        elif callable or allows_any_call:
            self.is_callable()
        else:
            self._callable = None
        self._expected_call_order = None

    def __getattribute__(self, name):
        """Favors stubbed out attributes, falls back to real attributes"""

        # this getter circumvents infinite loops:
        def g(n):
            return object.__getattribute__(self, n)

        if name in g("_declared_calls"):
            # if it's a call that has been declared
            # as that of the real object then hand it over:
            return g("_declared_calls")[name]
        elif name in g("_attributes"):
            # return attribute declared on real object
            return g("_attributes")[name]
        elif name in g("_properties"):
            # execute function and return result
            return g("_properties")[name]()
        else:
            # otherwise, first check if it's a call
            # of Fake itself (i.e. returns(),  with_args(), etc)
            try:
                self_call = g(name)
            except AttributeError:
                pass
            else:
                return self_call

            if g("_is_a_stub"):
                # Lazily create a attribute (which might later get called):
                stub = Fake(name=self._endpoint_name(name)).is_a_stub()
                self.has_attr(**{name: stub})
                return getattr(self, name)

            raise AttributeError(
                "%s object does not allow call or attribute '%s' "
                "(maybe you want %s.is_a_stub() ?)" % (self, name, self.__class__.__name__)
            )

    def __call__(self, *args, **kwargs):
        if "__init__" in self._declared_calls:
            # special case, simulation of __init__():
            call = self._declared_calls["__init__"]
            result = call(*args, **kwargs)
            if result is None:
                # assume more calls were expected / provided by the same fake
                return self
            else:
                # a new custom object has been declared
                return result
        elif self._callable:
            return self._callable(*args, **kwargs)
        elif self._is_a_stub:
            self.is_callable().returns_fake().is_a_stub()
            return self.__call__(*args, **kwargs)
        else:
            raise RuntimeError(
                "%s object cannot be called (maybe you want "
                "%s.is_callable() ?)" % (self, self.__class__.__name__)
            )

    def __setattr__(self, name, val):
        if hasattr(self, "_attributes") and name in self._attributes:
            self._attributes[name] = val
        else:
            object.__setattr__(self, name, val)

    def __repr__(self):
        return "fake:%s" % (self._name or "unnamed")

    def _declare_call(self, call_name, call):
        self._declared_calls[call_name] = call

    _assignment = re.compile(r"\s*(?P<name>[a-zA-Z0-9_]+)\s*=\s*(fudge\.)?Fake\(.*")

    def _guess_asn_from_file(self, frame):
        if frame.f_code.co_filename:
            if os.path.exists(frame.f_code.co_filename):
                cofile = open(frame.f_code.co_filename, "r")
                try:
                    for ln, line in enumerate(cofile):
                        # I'm not sure why -1 is needed
                        if ln == frame.f_lineno - 1:
                            possible_asn = line
                            m = self._assignment.match(possible_asn)
                            if m:
                                return m.group("name")
                finally:
                    cofile.close()

    def _guess_name(self):
        if not hasattr(sys, "_getframe"):
            # Stackless?
            return None
        if sys.platform.startswith("java"):
            # frame objects are completely different,
            # not worth the hassle.
            return None

        # get frame where class was instantiated,
        #   my_obj = Fake()
        #   ^
        #   we want to set self._name = 'my_obj'
        frame = sys._getframe(2)
        if len(frame.f_code.co_varnames):
            # at the top-most frame:
            co_names = frame.f_code.co_varnames
        else:
            # any other frame:
            co_names = frame.f_code.co_names

        # find names that are not locals.
        # this probably indicates my_obj = ...

        candidates = [n for n in co_names if n not in frame.f_locals]
        if len(candidates) == 0:
            # the value was possibly queued for deref
            #   foo = 44
            #   foo = Fake()
            return self._guess_asn_from_file(frame)
        elif len(candidates) == 1:
            return candidates[0]
        else:
            # we are possibly halfway through a module
            # where not all names have been compiled
            return self._guess_asn_from_file(frame)

    def _get_current_call(self):
        if not self._last_declared_call_name:
            if not self._callable:
                raise FakeDeclarationError(
                    "Call to a method that expects a predefined call but no such call exists.  "
                    "Maybe you forgot expects('method') or provides('method') ?"
                )
            return self._callable.get_call_object()
        exp = self._declared_calls[self._last_declared_call_name].get_call_object()
        return exp

    def _endpoint_name(self, endpoint):
        p = [self._name or "unnamed"]
        if endpoint != self._name:
            p.append(str(endpoint))
        return ".".join(p)

    def expects_call(self):
        """The fake must be called.

        .. doctest::
            :hide:

            >>> import fudge
            >>> fudge.clear_expectations()
            >>> fudge.clear_calls()

        This is useful for when you stub out a function
        as opposed to a class.  For example::

            >>> import fudge
            >>> remove = fudge.Fake('os.remove').expects_call()
            >>> fudge.verify()
            Traceback (most recent call last):
            ...
            AssertionError: fake:os.remove() was not called

        .. doctest::
            :hide:

            >>> fudge.clear_expectations()

        """
        self._callable = ExpectedCall(self, call_name=self._name, callable=True)
        return self

    def is_callable(self):
        """The fake can be called.

        This is useful for when you stub out a function
        as opposed to a class.  For example::

            >>> import fudge
            >>> remove = Fake('os.remove').is_callable()
            >>> remove('some/path')

        """
        self._callable = Call(self, call_name=self._name, callable=True)
        return self

    def is_a_stub(self):
        """Turns this fake into a stub.

        When a stub, any method is allowed to be called on the Fake() instance
        and any attribute can be accessed.  When an unknown attribute or
        call is made, a new Fake() is returned.  You can of course override
        any of this with :meth:`Fake.expects` and the other methods.
        """
        self._is_a_stub = True
        return self

    def calls(self, call):
        """Redefine a call.

        The fake method will execute your function.  I.E.::

            >>> f = Fake().provides('hello').calls(lambda: 'Why, hello there')
            >>> f.hello()
            'Why, hello there'

        """
        exp = self._get_current_call()
        exp.call_replacement = call
        return self

    def expects(self, call_name):
        """Expect a call.

        .. doctest::
            :hide:

            >>> import fudge
            >>> fudge.clear_expectations()
            >>> fudge.clear_calls()

        If the method *call_name* is never called, then raise an error.  I.E.::

            >>> session = Fake('session').expects('open').expects('close')
            >>> session.open()
            >>> fudge.verify()
            Traceback (most recent call last):
            ...
            AssertionError: fake:session.close() was not called

        .. note::
            If you want to also verify the order these calls are made in,
            use :func:`fudge.Fake.remember_order`.  When using :func:`fudge.Fake.next_call`
            after ``expects(...)``, each new call will be part of the expected order

        Declaring ``expects()`` multiple times is the same as
        declaring :func:`fudge.Fake.next_call`
        """
        if call_name in self._declared_calls:
            return self.next_call(for_method=call_name)

        self._last_declared_call_name = call_name
        c = ExpectedCall(self, call_name, call_order=self._expected_call_order)
        self._declare_call(call_name, c)
        return self

    def has_attr(self, **attributes):
        """Sets available attributes.

        I.E.::

            >>> User = Fake('User').provides('__init__').has_attr(name='Harry')
            >>> user = User()
            >>> user.name
            'Harry'

        """
        self._attributes.update(attributes)
        return self

    def has_property(self, **properties):
        """Sets available properties.

        I.E.::

            >>> mock_name = Fake().is_callable().returns('Jim Bob')
            >>> mock_age = Fake().is_callable().raises(AttributeError('DOB not set'))
            >>> user = Fake('User').has_property(name=mock_name, age=mock_age)
            >>> user.name
            'Jim Bob'
            >>> user.age
            Traceback (most recent call last):
            ...
            AttributeError: DOB not set

        """
        self._properties.update(properties)
        return self

    def next_call(self, for_method=None):
        """Start expecting or providing multiple calls.

        .. note:: next_call() cannot be used in combination with :func:`fudge.Fake.times_called`

        Up until calling this method, calls are infinite.

        For example, before next_call() ... ::

            >>> from fudge import Fake
            >>> f = Fake().provides('status').returns('Awake!')
            >>> f.status()
            'Awake!'
            >>> f.status()
            'Awake!'

        After next_call() ... ::

            >>> from fudge import Fake
            >>> f = Fake().provides('status').returns('Awake!')
            >>> f = f.next_call().returns('Asleep')
            >>> f = f.next_call().returns('Dreaming')
            >>> f.status()
            'Awake!'
            >>> f.status()
            'Asleep'
            >>> f.status()
            'Dreaming'
            >>> f.status()
            Traceback (most recent call last):
            ...
            AssertionError: This attribute of fake:unnamed can only be called 3 time(s).
            Call reset() if necessary or fudge.clear_calls().

        If you need to affect the next call of something other than the last declared call,
        use ``next_call(for_method="other_call")``.  Here is an example using getters and setters
        on a session object ::

            >>> from fudge import Fake
            >>> sess = Fake('session').provides('get_count').returns(1)
            >>> sess = sess.provides('set_count').with_args(5)

        Now go back and adjust return values for get_count() ::

            >>> sess = sess.next_call(for_method='get_count').returns(5)

        This allows these calls to be made ::

            >>> sess.get_count()
            1
            >>> sess.set_count(5)
            >>> sess.get_count()
            5

        When using :func:`fudge.Fake.remember_order` in combination with :func:`fudge.Fake.expects`
        and :func:`fudge.Fake.next_call` each new call will be part of the expected order.

        """
        last_call_name = self._last_declared_call_name
        if for_method:
            if for_method not in self._declared_calls:
                raise FakeDeclarationError(
                    "next_call(for_method=%r) is not possible; "
                    "declare expects(%r) or provides(%r) first"
                    % (for_method, for_method, for_method)
                )
            else:
                # set this for the local function:
                last_call_name = for_method
                # reset this for subsequent methods:
                self._last_declared_call_name = last_call_name

        if last_call_name:
            exp = self._declared_calls[last_call_name]
        elif self._callable:
            exp = self._callable
        else:
            raise FakeDeclarationError(
                "next_call() must follow provides(), " "expects() or is_callable()"
            )

        if getattr(exp, "expected_times_called", None) is not None:
            raise FakeDeclarationError("Cannot use next_call() in combination with times_called()")

        if not isinstance(exp, CallStack):
            # lazily create a stack with the last defined
            # expected call as the first on the stack:
            stack = CallStack(
                self,
                initial_calls=[exp],
                expected=isinstance(exp, ExpectedCall),
                call_name=exp.call_name,
            )

            # replace the old call obj using the same name:
            if last_call_name:
                self._declare_call(last_call_name, stack)
            elif self._callable:
                self._callable = stack
        else:
            stack = exp

        # hmm, we need a copy here so that the last call
        # falls off the stack.
        if stack.expected:
            next_call = ExpectedCall(
                self, call_name=exp.call_name, call_order=self._expected_call_order
            )
        else:
            next_call = Call(self, call_name=exp.call_name)
        stack.add_call(next_call)
        return self

    def provides(self, call_name):
        """Provide a call.

        The call acts as a stub -- no error is raised if it is not called.::

            >>> session = Fake('session').provides('open').provides('close')
            >>> import fudge
            >>> fudge.clear_expectations() # from any previously declared fakes
            >>> fudge.clear_calls()
            >>> session.open()
            >>> fudge.verify() # close() not called but no error

        Declaring ``provides()`` multiple times is the same as
        declaring :func:`fudge.Fake.next_call`

        """
        if call_name in self._declared_calls:
            return self.next_call(for_method=call_name)

        self._last_declared_call_name = call_name
        c = Call(self, call_name)
        self._declare_call(call_name, c)
        return self

    def raises(self, exc):
        """Set last call to raise an exception class or instance.

        For example::

            >>> import fudge
            >>> exc = ValueError("not enough parameters for insert")
            >>> db = fudge.Fake('db').provides('insert').raises(exc)
            >>> db.insert()
            Traceback (most recent call last):
            ...
            ValueError: not enough parameters for insert

        """
        exp = self._get_current_call()
        exp.exception_to_raise = exc
        return self

    def remember_order(self):
        """Verify that subsequent :func:`fudge.Fake.expects` are called in the right order.

        For example::

            >>> import fudge
            >>> db = fudge.Fake('db').remember_order().expects('insert').expects('update')
            >>> db.update()
            Traceback (most recent call last):
            ...
            AssertionError: Call #1 was fake:db.update();
            Expected: #1 fake:db.insert(), #2 fake:db.update(), end
            >>> fudge.clear_expectations()

        When declaring multiple calls using :func:`fudge.Fake.next_call`,
        each subsequent call will be added to the expected order of calls ::

            >>> import fudge
            >>> sess = fudge.Fake("session").remember_order().expects("get_id").returns(1)
            >>> sess = sess.expects("set_id").with_args(5)
            >>> sess = sess.next_call(for_method="get_id").returns(5)

        Multiple calls to ``get_id()`` are now expected ::

            >>> sess.get_id()
            1
            >>> sess.set_id(5)
            >>> sess.get_id()
            5
            >>> fudge.verify()
            >>> fudge.clear_expectations()

        """
        if self._callable:
            raise FakeDeclarationError(
                "remember_order() cannot be used for Fake(callable=True) or Fake(expect_call=True)"
            )
        self._expected_call_order = ExpectedCallOrder(self)
        registry.remember_expected_call_order(self._expected_call_order)
        return self

    def returns(self, val):
        """Set the last call to return a value.

        Set a static value to return when a method is called.  I.E.::

            >>> f = Fake().provides('get_number').returns(64)
            >>> f.get_number()
            64

        """
        exp = self._get_current_call()
        exp.return_val = val
        return self

    def returns_fake(self, *args, **kwargs):
        """Set the last call to return a new :class:`fudge.Fake`.

        Any given arguments are passed to the :class:`fudge.Fake` constructor

        Take note that this is different from the cascading nature of
        other methods.  This will return an instance of the *new* Fake,
        not self, so you should be careful to store its return value in a new
        variable.

        I.E.::

            >>> session = Fake('session')
            >>> query = session.provides('query').returns_fake(name="Query")
            >>> assert query is not session
            >>> query = query.provides('one').returns(['object'])

            >>> session.query().one()
            ['object']

        """
        exp = self._get_current_call()
        endpoint = kwargs.get("name", exp.call_name)
        name = self._endpoint_name(endpoint)
        kwargs["name"] = "%s()" % name
        fake = self.__class__(*args, **kwargs)
        exp.return_val = fake
        return fake

    def times_called(self, n):
        """Set the number of times an object can be called.

        When working with provided calls, you'll only see an
        error if the expected call count is exceeded ::

            >>> auth = Fake('auth').provides('login').times_called(1)
            >>> auth.login()
            >>> auth.login()
            Traceback (most recent call last):
            ...
            AssertionError: fake:auth.login() was called 2 time(s). Expected 1.

        When working with expected calls, you'll see an error if
        the call count is never met ::

            >>> import fudge
            >>> auth = fudge.Fake('auth').expects('login').times_called(2)
            >>> auth.login()
            >>> fudge.verify()
            Traceback (most recent call last):
            ...
            AssertionError: fake:auth.login() was called 1 time(s). Expected 2.

        .. note:: This cannot be used in combination with :func:`fudge.Fake.next_call`

        """
        if self._last_declared_call_name:
            actual_last_call = self._declared_calls[self._last_declared_call_name]
            if isinstance(actual_last_call, CallStack):
                raise FakeDeclarationError(
                    "Cannot use times_called() in combination with next_call()"
                )
        # else: # self._callable is in effect

        exp = self._get_current_call()
        exp.expected_times_called = n
        return self

    def with_args(self, *args, **kwargs):
        """Set the last call to expect specific argument values.

        The app under test must send all declared arguments and keyword arguments
        otherwise your test will raise an AssertionError.  For example:

        .. doctest::

            >>> import fudge
            >>> counter = fudge.Fake('counter').expects('increment').with_args(25, table='hits')
            >>> counter.increment(24, table='clicks')
            Traceback (most recent call last):
            ...
            AssertionError: fake:counter.increment(25, table='hits')
            was called unexpectedly with args (24, table='clicks')

        If you need to work with dynamic argument values
        consider using :func:`fudge.Fake.with_matching_args` to make looser declarations.
        You can also use :mod:`fudge.inspector` functions.  Here is an example of providing
        a more flexible ``with_args()`` declaration using inspectors:

        .. doctest::
            :hide:

            >>> fudge.clear_expectations()

        .. doctest::

            >>> import fudge
            >>> from fudge.inspector import arg
            >>> counter = fudge.Fake('counter')
            >>> counter = counter.expects('increment').with_args(
            ...                                         arg.any(),
            ...                                         table=arg.endswith("hits"))
            ...

        The above declaration would allow you to call counter like this:

        .. doctest::

            >>> counter.increment(999, table="image_hits")
            >>> fudge.verify()

        .. doctest::
            :hide:

            >>> fudge.clear_calls()

        Or like this:

        .. doctest::

            >>> counter.increment(22, table="user_profile_hits")
            >>> fudge.verify()

        .. doctest::
            :hide:

            >>> fudge.clear_expectations()

        """
        exp = self._get_current_call()
        if args:
            exp.expected_args = args
        if kwargs:
            exp.expected_kwargs = kwargs
        return self

    def with_matching_args(self, *args, **kwargs):
        """Set the last call to expect specific argument values if those arguments exist.

        Unlike :func:`fudge.Fake.with_args` use this if you want to only declare
        expectations about matching arguments.  Any unknown keyword arguments
        used by the app under test will be allowed.

        For example, you can declare positional arguments but ignore keyword arguments:

        .. doctest::

            >>> import fudge
            >>> db = fudge.Fake('db').expects('transaction').with_matching_args('insert')

        With this declaration, any keyword argument is allowed:

        .. doctest::

            >>> db.transaction('insert', isolation_level='lock')
            >>> db.transaction('insert', isolation_level='shared')
            >>> db.transaction('insert', retry_on_error=True)

        .. doctest::
            :hide:

            >>> fudge.clear_expectations()

        .. note::

            you may get more mileage out of :mod:`fudge.inspector` functions as
            described in :func:`fudge.Fake.with_args`

        """
        exp = self._get_current_call()
        if args:
            exp.expected_matching_args = args
        if kwargs:
            exp.expected_matching_kwargs = kwargs
        return self

    def without_args(self, *args, **kwargs):
        """Set the last call to expect that certain arguments will not exist.

        This is the opposite of :func:`fudge.Fake.with_matching_args`.  It will
        fail if any of the arguments are passed.

        .. doctest::

            >>> import fudge
            >>> query = fudge.Fake('query').expects_call().without_args(
            ...     'http://example.com', name="Steve"
            ... )

            >>> query('http://python.org', name="Joe")
            >>> query('http://example.com')
            Traceback (most recent call last):
            ...
            AssertionError: fake:query() was called unexpectedly with arg http://example.com
            >>> query("Joe", "Frank", "Bartholomew", "Steve")
            >>> query(name='Steve')
            Traceback (most recent call last):
            ...
            AssertionError: fake:query() was called unexpectedly with kwarg name=Steve
            >>> query('http://python.org', name='Steve')
            Traceback (most recent call last):
            ...
            AssertionError: fake:query() was called unexpectedly with kwarg name=Steve
            >>> query(city='Chicago', name='Steve')
            Traceback (most recent call last):
            ...
            AssertionError: fake:query() was called unexpectedly with kwarg name=Steve

            >>> query.expects_call().without_args('http://example2.com')
            fake:query
            >>> query('foobar')
            >>> query('foobar', 'http://example2.com')
            Traceback (most recent call last):
            ...
            AssertionError: fake:query() was called unexpectedly with arg http://example2.com

            >>> query.expects_call().without_args(name="Hieronymus")
            fake:query
            >>> query("Gottfried", "Hieronymus")
            >>> query(name="Wexter", other_name="Hieronymus")
            >>> query('asdf', name="Hieronymus")
            Traceback (most recent call last):
            ...
            AssertionError: fake:query() was called unexpectedly with kwarg name=Hieronymus
            >>> query(name="Hieronymus")
            Traceback (most recent call last):
            ...
            AssertionError: fake:query() was called unexpectedly with kwarg name=Hieronymus

            >>> query = fudge.Fake('query').expects_call().without_args(
            ...     'http://example.com', name="Steve"
            ... ).with_args('dog')
            >>> query('dog')
            >>> query('dog', 'http://example.com')
            Traceback (most recent call last):
            ...
            AssertionError: fake:query('dog') was called unexpectedly
            with args ('dog', 'http://example.com')
            >>> query()
            Traceback (most recent call last):
            ...
            AssertionError: fake:query('dog') was called unexpectedly with args ()

        """
        exp = self._get_current_call()
        if args:
            exp.unexpected_args = args
        if kwargs:
            exp.unexpected_kwargs = kwargs
        return self

    def with_arg_count(self, count):
        """Set the last call to expect an exact argument count.

        I.E.::

            >>> auth = Fake('auth').provides('login').with_arg_count(2)
            >>> auth.login('joe_user') # forgot password
            Traceback (most recent call last):
            ...
            AssertionError: fake:auth.login() was called with 1 arg(s) but expected 2

        """
        exp = self._get_current_call()
        exp.expected_arg_count = count
        return self

    def with_kwarg_count(self, count):
        """Set the last call to expect an exact count of keyword arguments.

        I.E.::

            >>> auth = Fake('auth').provides('login').with_kwarg_count(2)
            >>> auth.login(username='joe') # forgot password=
            Traceback (most recent call last):
            ...
            AssertionError: fake:auth.login() was called with 1 keyword arg(s) but expected 2

        """
        exp = self._get_current_call()
        exp.expected_kwarg_count = count
        return self



import inspect
import unittest
import fudge


class Freddie(object):
    pass


class TestPatch(unittest.TestCase):

    def setUp(self):
        fudge.clear_expectations()

    def test_decorator_on_def(self):
        class holder:
            test_called = False

        @fudge.patch("shutil.copy")
        def some_test(copy):
            import shutil

            holder.test_called = True
            assert isinstance(copy, fudge.Fake)
            self.assertEqual(copy, shutil.copy)

        self.assertEqual(some_test.__name__, "some_test")
        some_test()
        self.assertEqual(holder.test_called, True)
        import shutil

        assert not isinstance(shutil.copy, fudge.Fake)

    def test_decorator_on_class(self):
        class holder:
            test_called = False

        class MyTest(object):
            @fudge.patch("shutil.copy")
            def some_test(self, copy):
                import shutil

                holder.test_called = True
                assert isinstance(copy, fudge.Fake)
                assert copy == shutil.copy

        self.assertEqual(MyTest.some_test.__name__, "some_test")
        m = MyTest()
        m.some_test()
        self.assertEqual(holder.test_called, True)
        import shutil

        assert not isinstance(shutil.copy, fudge.Fake)

    def test_patch_many(self):
        class holder:
            test_called = False

        @fudge.patch("shutil.copy", "os.remove")
        def some_test(copy, remove):
            import shutil
            import os

            holder.test_called = True
            assert isinstance(copy, fudge.Fake)
            assert isinstance(remove, fudge.Fake)
            self.assertEqual(copy, shutil.copy)
            self.assertEqual(remove, os.remove)

        self.assertEqual(some_test.__name__, "some_test")
        some_test()
        self.assertEqual(holder.test_called, True)
        import shutil

        assert not isinstance(shutil.copy, fudge.Fake)
        import os

        assert not isinstance(os.remove, fudge.Fake)

    def test_with_patch(self):

        class holder:
            test_called = False

        def run_test():
            with fudge.patch("shutil.copy") as copy:
                import shutil

                assert isinstance(copy, fudge.Fake)
                self.assertEqual(copy, shutil.copy)
                holder.test_called = True

        run_test()
        self.assertEqual(holder.test_called, True)
        import shutil

        assert not isinstance(shutil.copy, fudge.Fake)

    def test_with_multiple_patches(self):

        class holder:
            test_called = False

        def run_test():
            with fudge.patch("shutil.copy", "os.remove") as fakes:
                copy, remove = fakes
                import shutil
                import os

                assert isinstance(copy, fudge.Fake)
                assert isinstance(remove, fudge.Fake)
                self.assertEqual(copy, shutil.copy)
                self.assertEqual(remove, os.remove)
                holder.test_called = True

        run_test()
        self.assertEqual(holder.test_called, True)
        import shutil

        assert not isinstance(shutil.copy, fudge.Fake)
        import os

        assert not isinstance(os.remove, fudge.Fake)

    def test_class_method_path(self):
        class ctx:
            sendmail = None

        @fudge.patch("smtplib.SMTP.sendmail")
        def test(fake_sendmail):
            import smtplib

            s = smtplib.SMTP()
            ctx.sendmail = s.sendmail

        test()
        assert isinstance(ctx.sendmail, fudge.Fake)
        import smtplib

        s = smtplib.SMTP()
        assert not isinstance(s.sendmail, fudge.Fake)

    def test_patch_obj(self):
        class holder:
            exc = Exception()

        patched = fudge.patch_object(holder, "exc", Freddie())
        self.assertEqual(type(holder.exc), type(Freddie()))
        patched.restore()
        self.assertEqual(type(holder.exc), type(Exception()))

    def test_patch_path(self):
        from os.path import join as orig_join

        patched = fudge.patch_object("os.path", "join", Freddie())
        import os.path

        self.assertEqual(type(os.path.join), type(Freddie()))
        patched.restore()
        self.assertEqual(type(os.path.join), type(orig_join))

    def test_patch_builtin(self):
        import datetime

        orig_datetime = datetime.datetime
        now = datetime.datetime(2010, 11, 4, 8, 19, 11, 28778)
        fake = fudge.Fake("now").is_callable().returns(now)
        patched = fudge.patch_object(datetime.datetime, "now", fake)
        try:
            self.assertEqual(datetime.datetime.now(), now)
        finally:
            patched.restore()
        self.assertEqual(datetime.datetime.now, orig_datetime.now)

    def test_patch_long_path(self):
        import fudge.tests._for_patch

        orig = fudge.tests._for_patch.some_object.inner
        long_path = "fudge.tests._for_patch.some_object.inner"
        with fudge.patch(long_path) as fake:
            assert isinstance(fake, fudge.Fake)
        self.assertEqual(fudge.tests._for_patch.some_object.inner, orig)

    def test_patch_non_existant_path(self):
        with self.assertRaises(ImportError):
            with fudge.patch("__not_a_real_import_path.nested.one.two.three"):
                pass

    def test_patch_non_existant_attribute(self):
        with self.assertRaises(AttributeError):
            with fudge.patch("fudge.tests._for_patch.does.not.exist"):
                pass

    def test_patch_builtin_as_string(self):
        import datetime

        orig_datetime = datetime.datetime
        now = datetime.datetime(2006, 11, 4, 8, 19, 11, 28778)
        fake_dt = fudge.Fake("datetime").provides("now").returns(now)
        patched = fudge.patch_object("datetime", "datetime", fake_dt)
        try:
            # timetuple is a workaround for strange Jython behavior!
            self.assertEqual(datetime.datetime.now().timetuple(), now.timetuple())
        finally:
            patched.restore()
        self.assertEqual(datetime.datetime.now, orig_datetime.now)

    def test_patched_context(self):
        class Boo:
            fargo = "is over there"

        ctx = fudge.patched_context(Boo, "fargo", "is right here")
        # simulate with fudge.patched_context():
        ctx.__enter__()
        self.assertEqual(Boo.fargo, "is right here")
        ctx.__exit__(None, None, None)
        self.assertEqual(Boo.fargo, "is over there")

    def test_base_class_attribute(self):
        class Base(object):
            foo = "bar"

        class Main(Base):
            pass

        fake = fudge.Fake()
        p = fudge.patch_object(Main, "foo", fake)
        self.assertEqual(Main.foo, fake)
        self.assertEqual(Base.foo, "bar")
        p.restore()
        self.assertEqual(Main.foo, "bar")
        assert "foo" not in Main.__dict__, "Main.foo was not restored correctly"

    def test_bound_methods(self):
        class Klass(object):
            def method(self):
                return "foozilate"

        instance = Klass()
        fake = fudge.Fake()
        p = fudge.patch_object(instance, "method", fake)
        self.assertEqual(instance.method, fake)
        p.restore()
        self.assertEqual(instance.method(), Klass().method())
        assert inspect.ismethod(instance.method)
        assert "method" not in instance.__dict__, "instance.method was not restored correctly"

    def test_staticmethod_descriptor(self):
        class Klass(object):
            @staticmethod
            def static():
                return "OK"

        fake = fudge.Fake()
        p = fudge.patch_object(Klass, "static", fake)
        self.assertEqual(Klass.static, fake)
        p.restore()
        self.assertEqual(Klass.static(), "OK")

    def test_property(self):
        class Klass(object):
            @property
            def prop(self):
                return "OK"

        exact_prop = Klass.prop
        instance = Klass()
        fake = fudge.Fake()
        p = fudge.patch_object(instance, "prop", fake)
        self.assertEqual(instance.prop, fake)
        p.restore()
        self.assertEqual(instance.prop, "OK")
        self.assertEqual(Klass.prop, exact_prop)

    def test_inherited_property(self):
        class SubKlass(object):
            @property
            def prop(self):
                return "OK"

        class Klass(SubKlass):
            pass

        exact_prop = SubKlass.prop
        instance = Klass()
        fake = fudge.Fake()
        p = fudge.patch_object(instance, "prop", fake)
        self.assertEqual(instance.prop, fake)
        p.restore()
        self.assertEqual(instance.prop, "OK")
        assert "prop" not in Klass.__dict__, "Klass.prop was not restored properly"
        self.assertEqual(SubKlass.prop, exact_prop)
